"""
#############################################################################

Copyright (C) 2024, Michele Cappellari
E-mail: michele.cappellari_at_physics.ox.ac.uk

Updated versions of the software are available from my web page
http://purl.org/cappellari/software

If you have found this software useful for your research,
I would appreciate an acknowledgement to the use of the
"JAM modelling package of Cappellari (2008)".

https://ui.adsabs.harvard.edu/abs/2008MNRAS.390...71C

This software is provided as is without any warranty whatsoever.
Permission to use, for non-commercial purposes is granted.
Permission to modify for personal or internal use is granted,
provided this copyright and disclaimer are included unchanged
at the beginning of the file. All other rights are reserved.
In particular, redistribution of the code is not allowed.

#############################################################################

MODIFICATION HISTORY:
    V1.0.0: Written and tested. Michele Cappellari, Oxford, 4 February 2016
    V1.0.1: Written documentation. MC, Oxford, 4 April 2024

"""
import numpy as np
from scipy import special

from jampy.quad1d import quad1d

###################################################################################

def mge_cylindrical_mass(surf, sigma, qobs, rad, distance=1e-5):
    """
    Calculates the mass of a projected axisymmetric MGE enclosed inside a
    cylinder of radius R with symmetry axis along the line-of-sight.

    Input Parameters
    ----------------
    
    surf: array_like with shape (n,)
        Peak value of every Gaussian in the `Multi-Gaussian Expansion
        <https://pypi.org/project/mgefit/>`_ (MGE) describing the surface
        brightness in units of ``Lsun/pc^2`` (solar luminosities per
        ``parsec^2``).
    sigma: array_like with shape (n,)
        Dispersion (sigma) in arcsec of the MGE Gaussians.
    qobs: array_like with shape (n,)
        Observed axial ratio (q') of the MGE Gaussians.
    rad: array_like with shape (m,)
        Radii in arcsec at which one wants to compute the enclosed mass.
    
    Optional Keywords
    -----------------

    distance: float 
        distance in Mpc (default distance = 10pc = 1e-5Mpc)

    Output Parameters
    -----------------

    Returns the projected mass in solar masses enclosed within the m radii.

    """
    rad = np.atleast_1d(rad)
    pc = distance*np.pi/0.648  # Constant factor to convert arcsec --> pc
    lum = 2*np.pi*surf*qobs*(sigma*pc)**2  # total luminosity each Gaussian

    def integrand(rad):
        h = rad[:, None]/(2*qobs*sigma)
        mass_R = 2*lum*h*np.exp(-2*(h*qobs)**2)*special.i0e(h**2*(qobs**2 - 1))/sigma
        return mass_R.sum(1)

    return [quad1d(integrand, [0, r], singular=True).integ for r in rad]

###################################################################################



"""
    Copyright (C) 2022-2024, Michele Cappellari

    E-mail: michele.cappellari_at_physics.ox.ac.uk

    Updated versions of the software are available from my web page
    https://purl.org/cappellari/software

    This software is provided as is without any warranty whatsoever.
    Permission to use, for non-commercial purposes is granted.
    Permission to modify for personal or internal use is granted,
    provided this copyright and disclaimer are included unchanged
    at the beginning of the file. All other rights are reserved.
    In particular, redistribution of the code is not allowed.

    Changelog
    ---------

    V1.0.0: Michele Cappellari, Oxford, 17 January 2022
    +++++++++++++++++++++++++++++++++++++++++++++++++++

    - Written and tested as a separate procedure.

    Vx.x.x: MC, Oxford
    ++++++++++++++++++

    - Additional changes are documented in the CHANGELOG of the JamPy package.


"""

import numpy as np
import matplotlib.pyplot as plt

from mgefit.mge_fit_1d import mge_fit_1d
from jampy.jam_axi_proj import jam_axi_proj, rotate_points
from jampy.cosmology_distance import angular_diameter_distance
from plotbin.plot_velfield import plot_velfield

###############################################################################

def sersic_profile(n, rad):
    """Sersic profile with rad=R/Re"""

    # This is eq.(28) of Cappellari (2023, MNRAS)
    # It has a maximum absolute error of 5.6e-4 in the interval n = [0.2, 16]
    bn = 0.000207/n**2 + 0.015987/n - 0.34025 + 2.0015*n
    surf = np.exp(-bn*(rad)**(1/n))   # r^(1/n) Sersic profile

    return surf  # Profile is NOT normalized

###############################################################################

def sersic_mge(n_ser, ngauss, lg_rmax, plot, quiet):
    """MGE for a Sersic profile"""

    m = 300 # Number of values to sample the profile for the fit
    r = np.logspace(-lg_rmax, lg_rmax, m)
    rho = sersic_profile(n_ser, r)
    w = rho > rho[0]/1e40  # samples at most 40 orders of magnitude in density
    total_counts, sigma = mge_fit_1d(r[w], rho[w], ngauss=ngauss, plot=plot,
                                     rbounds=[np.min(r)/2, np.max(r)], quiet=quiet).sol
    surf = total_counts/(np.sqrt(2*np.pi)*sigma)  # Surface density in Msun/pc**2

    if plot:
        plt.pause(1)

    return surf, sigma

###############################################################################

class jam_axi_sersic_mass:
    """
    jam_axi_sersic_mass
    ===================

    Purpose
    -------

    Compute the mass ``M_Ser`` and uncertainty of a mass-follows-light Sersic
    model, at the given distance, that produces a given second moment of the
    stellar velocity ``sigma_ap`` inside a rectangular aperture of sides
    ``dxy_ap``, assuming an intrinsic axial ratio ``qintr``.

    I describe this procedure in Sec.5.2 of 
    `Cappellari (2023) <https://ui.adsabs.harvard.edu/abs/2023MNRAS.526.3273C>`_.

    Ideally, the Sersic model should be fitted to the photometry in a band close
    to the wavelength of the spectroscopic observations used to extract ``sigma_ap``.

    The physical meaning of the output ``M_Ser`` is the following: when divided by
    the total luminosity ``L_Ser`` of the Sersic model in a given band, then::

        (M/L)_ap = M_Ser/L_Ser

    closely approximates the average mass-to-light ratio in the same band,
    within a sphere of radius ``R_ap`` centred on the galaxy nucleus (see Fig.9 of 
    `Cappellari et al. 2013 <https://ui.adsabs.harvard.edu/abs/2013MNRAS.432.1709C>`_).

    This procedure is meant as a simple replacement for the less accurate
    virial mass estimators. This is to be used when one has a single
    determination of sigma for a galaxy, rather than spatially-resolved data.

    Calling Sequence
    ----------------

    .. code-block:: python

        from jampy.jam_axi_sersic import jam_axi_sersic_mass

        jam = jam_axi_sersic_mass(re_maj_ser, n_ser, qobs, qintr, sigma_ap,
                sigma_ap_err, dxy_ap, sigma_psf, distance, angle=0., beta=0., 
                ngauss=20, lg_rmax=2., plot=True, quiet=False)
        
        print(f"Sersic Mass (M_Sun): {10**jam.lg_mjam:.3g}")

    Input Parameters
    ----------------

    re_maj_ser: float
        Major axis of the ellipse containing half of the total luminosity of the
        Sersic (1968) model which describes the surface brightness of the galaxy
        under study::
            
            surf(m) = exp[-bn*m^(1/n_ser)]
            m^2 = x^2 + (y/qobs)^2

    n_ser: float
        Sersic index. ``n_ser=1`` corresponds to an exponential profile, while
        ``n_ser=4`` represents a de Vaucouleurs (1948) profile.
    qobs: float
        The observed axial ratio of the Sersic model.
    qintr: float
        Assumed intrinsic axial ratio of the galaxy model.
    sigma_ap:
        Observed second velocity moment within a rectangular aperture. This is
        generally the velocity dispersion measured from a spectrum within the
        aperture.
    sigma_ap_err: float
        ``1sigma`` uncertainty on ``sigma_ap``.
    dxy_ap: array_like with shape (2,)
        Sizes of the rectangular aperture ``[dx, dy]``.
    sigma_psf: float
        Dispersion of the PSF.
    distance: float
        Adopted distance of the galaxy in Mpc

    Optional Keywords
    -----------------
    
    angle: float
        Angle in degrees between the first dimension of the rectangular
        aperture and the galaxy projected major axis.
    beta: float
        Adopted anisotropy ``beta = 1 - (sigma_z/sigma_R)^2``.
    lg_rmax: float
        The Sersic profile is fitted from ``1/10^lg_rmax`` to ``10^lg_rmax``.
    ngauss: int
        Number of Gaussians used to fit the Sersic model with an MGE.
    plot: bool
        Set to ``True`` to plot the model, for debugging purposes.
    quiet: bool
        Set to ``True`` to suppress printed output.
                
    Output Parameters
    -----------------

    Stored as attributes of the ``jam_axi_sersic_mass`` class.

    .lg_mjam: float
        Logarithm (base 10) of the total dynamical mass.
    .d_lg_mjam: float
        1sigma uncertainty on lg_mjam from error propagation, only considering
        the uncertainty in the input ``sigma_ap``.
    .mge_pot: array like with shape (3, ngauss)
        (surf, sigma, qobs) parameters of the MGE parameters describing the
        mass distribution parametrized by the Sersic profile.
    .sigma_e: float
        Aperture and seeing-corrected value of the luminosity-weighted second
        moment inside the half-light ellipse. This is computed from the JAM
        model and provides an estimate of the sigma_e that one could have
        measured from spatially-resolved high-resolution observations of the
        galaxy.
    """
    def __init__(self, re_maj_ser, n_ser, qobs, qintr, sigma_ap, sigma_ap_err,
                 dxy_ap, sigma_psf, distance, angle=0., beta=0., ngauss=20, 
                 lg_rmax=2., plot=True, quiet=False):

        assert qintr <= qobs, "Must be `qintr <= qobs`"
        inc = np.degrees(np.arctan2(np.sqrt(1 - qobs**2), np.sqrt(qobs**2 - qintr**2)))

        # Adopt as reference an arbitrary galaxy mass and find how much I need
        # to scale it to match the observed sigma
        mass = 1e11

        pc = distance*np.pi/0.648  # Factor to convert arcsec --> pc (with distance in Mpc)
        surf, sigma = sersic_mge(n_ser, ngauss, lg_rmax, plot, quiet)
        sigma = sigma*re_maj_ser
        qobs_mge = np.full_like(sigma, qobs)            # projected axial ratio

        mtot = 2*np.pi*(surf*qobs_mge*(sigma*pc)**2).sum()
        surf *= mass/mtot  # Rescale MGE to have input total mass

        dx, dy = dxy_ap
        npix = 100    # pixels discretizing the aperture
        x, xstep = np.linspace(-dx/2, dx/2, npix, retstep=True)      # Avoids (0, 0)
        y, ystep = np.linspace(-dy/2, dy/2, npix, retstep=True)      # Avoids (0, 0)
        pixsize = np.sqrt(xstep*ystep)
        xbin, ybin = map(np.ravel, np.meshgrid(x, y))
        xbin, ybin = rotate_points(xbin, ybin, angle)
        beta = np.full_like(sigma, beta)
        mbh = mass*0.005   # 0.5% BH mass (eq.11 of Kormendy & Ho 2013, ARA&A)

        jam = jam_axi_proj(surf, sigma, qobs_mge, surf, sigma, qobs_mge, inc, 
                           mbh, distance, xbin, ybin, beta=beta, step=pixsize,
                           sigmapsf=sigma_psf, quiet=quiet, pixsize=pixsize)

        sigma_ap2 = (jam.flux*jam.model**2).sum()/jam.flux.sum()
        scale = sigma_ap**2/sigma_ap2
        lg_mjam = np.log10(mass*scale)
        d_lg_mjam = 2*sigma_ap_err/(sigma_ap*np.log(10))   # error propagation

        if plot:
            plt.clf()        
            plot_velfield(xbin, ybin, jam.model, flux=jam.flux, nodots=1)
            plt.title("JAM $V_{\\rm rms}$")
            plt.pause(1)

        # Compute the effective velocity second moment sigma_e within the
        # half-light ellipse

        x = np.linspace(-re_maj_ser, re_maj_ser, npix)   # Avoids (0, 0)
        xbin, ybin = map(np.ravel, np.meshgrid(x, x*qobs))

        jam = jam_axi_proj(surf, sigma, qobs_mge, surf*scale, sigma, qobs_mge, 
                           inc, mbh, distance, xbin, ybin, beta=beta, quiet=quiet)

        w = xbin**2 + (ybin/qobs)**2 < re_maj_ser**2
        sigma_e2 = (jam.flux[w]*jam.model[w]**2).sum()/jam.flux[w].sum()

        if plot:
            plt.clf()        
            plot_velfield(xbin, ybin, jam.model, flux=jam.flux, nodots=1)
            plt.plot(xbin[~w], ybin[~w], '+')
            plt.title("JAM $V_{\\rm rms}$")
            plt.pause(1)

        self.lg_mjam = lg_mjam
        self.d_lg_mjam = d_lg_mjam
        self.mge_pot = surf*scale, sigma, qobs_mge  # mass MGE
        self.sigma_e = np.sqrt(sigma_e2)
        self.sigma_ap = sigma_ap

        if not quiet:
            print(f"Aperture {dx:.1f}x{dy:.1f} arcsec at angle = {angle:.0f} deg")
            print(f"Input sigma_ap within the aperture: ({sigma_ap:.0f} +/- {sigma_ap_err:.0f}) km/s")
            print(f"Corrected sigma_e within Re: {self.sigma_e:.3g} km/s")
            print(f'lg(M_JAM/M_Sun) = {lg_mjam:.2f} +/- {d_lg_mjam:.2f}')


###############################################################################

def jam_axi_sersic_mass_example():
    """Usage example for jam_axi_sersic_mass"""

    # Input parameters
    dxy_ap = [1.0, 0.5]     # sides of rectangular aperture in arcsec
    n_ser = 4.0             # Sersic exponent (e.g. n = 4 --> de Vaucouleurs)
    qobs = 0.7              # Observed axial ratio of the fitted Sersic model
    qintr = 0.4             # Assumed intrinsic axial ratio q = b/a
    beta = 0.2              # Anisotropy beta = 1 - (sig_z/sig_R)^2
    re_maj_ser = 2.0        # arcsec. Semimajor axis of half-light ellipse of Sersic model
    sigma_psf = 0.5/2.355   # Dispersion of the PSF in arcsec
    redshift = 0.8          # Galaxy redshift
    sigma_ap = 200.         # km/s. Observed second velocity moment (sigma)
    sigma_ap_err = 15       # km/s. 1sigma uncertainty on sigma_ap

    # Computation
    dist_ang = angular_diameter_distance(redshift)  # D_A angular diameter distance

    jam = jam_axi_sersic_mass(re_maj_ser, n_ser, qobs, qintr, sigma_ap, sigma_ap_err, 
                              dxy_ap, sigma_psf, dist_ang, angle=30, beta=beta)
    
    print(f"\nSersic Mass (M_Sun): {10**jam.lg_mjam:.3g}")

###############################################################################

if __name__ == '__main__':

    jam_axi_sersic_mass_example()



"""
    Copyright (C) 2007-2023, Michele Cappellari

    E-mail: michele.cappellari_at_physics.ox.ac.uk

    Updated versions of the software are available from PyPi
    https://pypi.org/project/jampy

    If you have found this software useful for your research,
    I would appreciate an acknowledgement and a link to the website.

    This software is provided as is without any warranty whatsoever.
    Permission to use, for non-commercial purposes is granted.
    Permission to modify for personal or internal use is granted,
    provided this copyright and disclaimer are included unchanged
    at the beginning of the file. All other rights are reserved.
    In particular, redistribution of the code is not allowed.

############################################################################


CHANGELOG
---------

    V1.0.0: Written and tested against the corresponding MATLAB version,
        Michele Cappellari, Oxford, 22 October, 2007

    V1.1.0: Allow function parameters to be passed via the FUNCTARGS keyword.
        MC, Oxford, 23 October 2007

    V1.1.1: Added STATUS keyword. Provide more informative error messages on failure.
        MC, Windoek, 5 October 2008

    V1.1.2: Renamed CAP_QUADVA to avoid potential naming conflicts.
        MC, Paranal, 8 November 2013

    V2.0.0: Translated from IDL into Python. MC, Paranal, 14 November 2013

    V2.0.1: Fixed possible program stop. MC, Oxford, 27 January 2014

    V2.0.2: Support both legacy Python 2.7 and Python 3. MC, Oxford, 25 May 2014

    V2.0.3: Dropped Python 2.7 support. MC, Oxford, 12 May 2018

    V2.1.0: Included optional plotting of sampled points. MC, Oxford, 25 June 2019

    V3.0.0:
        - Changed interface as a class and renamed quadva --> quad1d
        - Print number of function evaluation if ``verbose=1``.
        - New ``singular`` keyword: set ``singular=False`` to prevent
          transforming the coordinates of the integrand. This is useful
          when one wants to transform the coordinates outside ``quad1d``.
        MC, Oxford, 4 February 2022

    V3.1.0: Change defaults to `singular=False` and `epsabs=0` like `quad2d`.
        MC, Oxford, 3 January 2023

    V4.0.0: Allow for quadrature of vector functions. Uses the same set of
        evaluation points for all components of the function and ensures that
        the convergence criterion is satisfied for all.
        MC, Oxford, 11 June 2023

"""
import numpy as np
import matplotlib.pyplot as plt

EPS = 100*np.finfo(float).eps

################################################################################

def _qva_split(interval):
    """
    If breakpoints are specified, split sub-intervals in
    half as needed to get a minimum of 10 sub-intervals.
    """
    v = interval
    while True:
        npts = interval.size
        if npts > 10:
            break
        v = np.zeros(npts*2 - 1)
        v[0::2] = interval
        v[1::2] = interval[:-1] + 0.5*np.diff(interval)
        interval = v

    return v

################################################################################

def _qva_check_spacing(x):
    ax = np.abs(x)
    too_close = np.any(np.diff(x) <= EPS*np.maximum(ax[:-1], ax[1:]))

    return too_close

################################################################################

def _qva_f1(fun, t, a, b, args):
    """
    Transform to weaken singularities at both ends: [a, b] -> [-1, 1]
    """
    x = 0.25*(b - a)*t*(3. - t**2) + 0.5*(b + a)
    too_close = _qva_check_spacing(x)
    if too_close:
        y = []
    else:
        y = fun(x, *args)
        y *= 0.75*(b - a)*(1. - t**2)

    return y, too_close

################################################################################

def _qva_f2(fun, t, a, b, args):
    """
    Transform to weaken singularity at left end: [a, np.Inf) -> [0, Inf).
    Then transform to finite interval: [0, Inf) -> [0, 1].
    """
    Tt = t/(1. - t)
    x = a + Tt**2
    too_close = _qva_check_spacing(x)
    if too_close:
        y = []
    else:
        y = fun(x, *args)
        y *= 2.*Tt/(1. - t)**2

    return y, too_close

################################################################################

def _qva_f3(fun, t, a, b, args):
    """
    Transform to weaken singularity at right end: (-Inf, b] -> (-Inf, 0].
    Then transform to finite interval: (-Inf, 0] -> (-1, 0].
    """
    Tt = t/(1. + t)
    x = b - Tt**2
    too_close = _qva_check_spacing(x)
    if too_close:
        y = []
    else:
        y = fun(x, *args)
        y *= -2.*Tt/(1. + t)**2

    return y, too_close

################################################################################

def _qva_f4(fun, t, a, b, args):
    """
    Transform to finite interval: (-Inf, Inf) -> (-1, 1)
    """
    x = t/(1. - t**2)
    too_close = _qva_check_spacing(x)
    if too_close:
        y = []
    else:
        y = fun(x, *args)
        y *= (1. + t**2)/(1. - t**2)**2

    return y, too_close

################################################################################

def _qva_f5(fun, t, a, b, args):
    """
    Simple linear transformation: [a, b] -> [-1, 1]
    """
    x = 0.5*((b + a) + (b - a)*t)
    too_close = _qva_check_spacing(x)
    if too_close:
        y = []
    else:
        y = fun(x, *args)
        y *= 0.5*(b - a)

    return y, too_close

################################################################################

def _qva_Vadapt(f, tinterval, rtol, atol, samples, nodes, wt, ewt, fun, a, b, args, plot):
    """ Main loop of the quadrature """

    tbma = np.abs(tinterval[-1] - tinterval[0])  # length of transformed interval

    # Initialize array of subintervals of [a,b].
    subs = np.column_stack([tinterval[:-1], tinterval[1:]])  # Two columns array[n, 2]

    # Initialize partial sums.
    IfxOK = errOK = 0

    # Initialize main loop
    OK = first = converged = True
    Ifx = errbnd = fcall = vfcall = 0

    while True:
        # SUBS contains subintervals of [a,b] where the integral is not
        # sufficiently accurate.  The first row of SUBS holds the left end
        # points and the second row, the corresponding right end points.
        midpt = subs.sum(1)/2      # midpoints of the subintervals
        halfh = np.diff(subs, axis=1)/2     # half the lengths of the subintervals
        x = nodes*halfh + midpt[:, None]    # broadcasting midpt
        halfh = halfh.ravel()
        x = x.ravel()

        fx, too_close = f(fun, x, a, b, args)
        fcall += x.size
        vfcall += 1

        # Quit if mesh points are too close or too close to a
        # singular point or got into trouble on first evaluation.
        not_finite = np.any(np.isinf(fx))
        if too_close or not_finite:
            break

        if plot:
            if first:
                plt.figure('quad1d')
                plt.clf()
                plt.xlabel('x')
                plt.ylabel('y')
                xall, yall = x, fx
            else:
                yall = np.column_stack([yall, fx])
                xall = np.append(xall, x)
            plt.figure('quad1d')
            plt.plot(x, fx.T, '+')

        if fx.ndim == 1:
            fx = fx[None, :]

        fx = fx.reshape(len(fx), -1, samples)

        # Quantities for subintervals.
        Ifxsubs = fx @ wt * halfh
        errsubs = fx @ ewt * halfh

        # Quantities for all of [a, b].
        Ifx = Ifxsubs.sum(1) + IfxOK
        errbnd = abs(errsubs.sum(1) + errOK)

        # Test for convergence:
        tol = np.maximum(atol, rtol*np.abs(Ifx))
        if np.all(errbnd <= tol):
            converged = True
            break

        # Locate subintervals where the approximate integrals are
        # sufficiently accurate and use them to update partial sums.
        good = np.all(np.abs(errsubs) <= (2/tbma)*halfh*tol[:, None], 0)
        errOK += np.sum(errsubs[:, good])
        IfxOK += np.sum(Ifxsubs[:, good])
        # Only keep subintervals which are not yet sufficiently accurate
        subs = subs[~good, :]
        if subs.size == 0:  # all intervals are accurate
            converged = True
            break

        # Split the remaining subintervals in half. Quit if splitting
        # results in too many subintervals.
        many_subint = 2*subs.size > 650  # multiplied limit by 10x MC 26/FEB/2008
        if many_subint:
            break
        midpt = subs.sum(1)/2
        tmp = np.column_stack([subs[:, 0], midpt, midpt, subs[:, 1]])
        subs = tmp.reshape(-1, 2)  # ---> subs[n, 2]
        first = False

    if first and not converged:
        if too_close:
            print('***Sub intervals too close.')
        elif not_finite:
            print('***Infinite values in integrand.')
        elif many_subint:
            print('***Too many sub intervals.')
        OK = False

    if plot:
        plt.figure('quad1d')
        w = np.argsort(xall)
        plt.plot(xall[w], yall.T[w])

    return Ifx, errbnd, OK, fcall, vfcall

################################################################################

class quad1d:
    """
    Calling sequence: ``q = quad1d(fun, interval, epsrel=1e-5, epsabs=0)``
    
    ``q.integ`` and ``q.errbnd`` will contain the value of the integral 
    and the corresponding estimate of the error bound.

    When ``INTERVAL = [A,B]``, computes the integral of a continuous function
    ``f(x)`` from A to B for A < B.  A can be ``-np.inf`` and/or B can be
    ``np.inf``.

    FUN accepts a row vector X and returns a vector Y with ``Y[m] = f(X[m])``
    for ``m = 1,...,len(X)``.

    FUN can also be a vector function ``fun = [f_1, f_2,... f_n]``.
    In this case the FUN must return a ``matrix[j, :] = f_j(X)``.
    In this case the procedure returns the integral of all components,
    computed using the same set of evaluation points. The convergence
    criterion ensures that all components satisfy the required tolerance.
    
    quad1d returns an approximation Ifx to the integral and optionally an
    approximate bound, ERRBND, on the error ``|integral - Ifx|``. It attempts
    to compute Ifx such that ``|Ifx - integral| <= max(ABSTOL,RELTOL*|Ifx|)``.
    If quad1d is unsuccessful, Ifx and ERRBND are still meaningful, so a
    warning is issued that includes ERRBND.

    If the interval is infinite, say ``[a, np.inf)``, then for the integral to
    exist, ``f(x)`` must decay as ``x -> np.inf`` and quad1d requires it to
    decay rapidly. Special methods should be used for oscillatory functions on
    infinite intervals, but quad1d can be used if ``f(x)`` decays fast enough.

    quad1d will integrate ``f(x)`` that are singular at finite end points if
    the singularities are not too strong. For example, it will integrate
    ``f(x)`` that behave like ``log|x - c|`` or ``|x - c|^p`` for ``p >= -1/2``
    with ``c = a`` and/or ``c = b``. If ``f(x)`` is singular at points inside
    ``(A,B)``, write the integral as a sum of integrals over subintervals with
    the singular points as end points, compute them with quad1d, and add the
    results.

    quad1d starts with samples of ``f(x)`` at 150 points in ``(A,B)``. It must
    be able to recognize the behaviour of ``f(x)`` from these samples, so if
    ``f(x)`` oscillates very rapidly or has sharp peaks, it may be necessary to
    subdivide the interval. To do this, make INTERVAL an array with entries
    (breakpoints) that increase from ``A = INTERVAL[0]`` to ``B = INTERVAL[-1]``.
    If ``f(x)`` is only piecewise smooth, the points of discontinuity should
    be breakpoints.

    Based on the algorithm "Vectorized Adaptive Quadrature in Matlab"
    `L.F. Shampine, Journal of Computational and Applied Mathematics
    211 (2008) 131-140 <http://dx.doi.org/10.1016/j.cam.2006.11.021>`_

    """
    def __init__(self, fun, interval, epsrel=1e-5, epsabs=0, plot=False,
                 verbose=False, singular=False, args=()):

        interval = np.asarray(interval)
        nint = interval.size
        assert nint >= 2, "INTERVAL must be a real vector of at least two entries."
        assert np.all(np.diff(interval) > 0), "Entries of INTERVAL must strictly increase."
        assert epsrel > 0 or epsabs > 0, "Either `epsrel` or `epsabs` must be positive"

        a = interval[0]
        b = interval[-1]

        # Generally the error test is a mixed one, but pure absolute error
        # and pure relative error are allowed.  If a pure relative error
        # test is specified, the tolerance must be at least 100*EPS. 
        #
        rtol = max(epsrel, EPS)
        atol = epsabs

        # Gauss-Kronrod (7,15) pair. Use symmetry in defining nodes and weights.
        #
        samples = 15
        pnodes = np.array(
            [0.2077849550078985, 0.4058451513773972, 0.5860872354676911,
             0.7415311855993944, 0.8648644233597691, 0.9491079123427585,
             0.9914553711208126])
        nodes = np.hstack([-pnodes[::-1], 0, pnodes])
        pwt = np.array(
            [0.2044329400752989, 0.1903505780647854, 0.1690047266392679,
             0.1406532597155259, 0.1047900103222502, 0.06309209262997855,
             0.02293532201052922])
        wt = np.hstack([pwt[::-1], 0.2094821410847278, pwt])
        pwt7 = np.array([0, 0.3818300505051189, 0, 0.2797053914892767,
                         0, 0.1294849661688697, 0])
        ewt = wt - np.hstack([pwt7[::-1], 0.4179591836734694, pwt7])

        # Identify the task. If breakpoints are specified, work out
        # how they map into the standard interval.
        #
        if singular or np.inf in [a, b]:

            if a != -np.Inf and b != np.Inf:
                if nint > 2:
                    # Analytical transformation suggested by K.L. Metlov:
                    alpha = 2.*np.sin(np.arcsin((a + b - 2.*interval[1:-1])/(a - b))/3.)
                    tinterval = np.hstack([-1., alpha, 1.])
                    tinterval = _qva_split(tinterval)
                else:
                    tinterval = np.linspace(-1, 1, 11)

                Ifx, errbnd, OK, fcall, vfcall = _qva_Vadapt(
                    _qva_f1, tinterval, rtol, atol, samples, nodes, wt, ewt,
                    fun, a, b, args, plot)

            elif a != -np.Inf and b == np.Inf:
                if nint > 2:
                    alpha = np.sqrt(interval[1:-1] - a)
                    tinterval = np.hstack([0., alpha/(1. + alpha), 1.])
                    tinterval = _qva_split(tinterval)
                else:
                    tinterval = np.linspace(0, 1, 11)

                Ifx, errbnd, OK, fcall, vfcall = _qva_Vadapt(
                    _qva_f2, tinterval, rtol, atol, samples, nodes, wt, ewt,
                    fun, a, b, args, plot)

            elif a == -np.Inf and b != np.Inf:
                if nint > 2:
                    alpha = np.sqrt(b - interval[1:-1])
                    tinterval = np.hstack([-1., -alpha/(1. + alpha), 0.])
                    tinterval = _qva_split(tinterval)
                else:
                    tinterval = np.linspace(-1, 0, 11)

                Ifx, errbnd, OK, fcall, vfcall = _qva_Vadapt(
                    _qva_f3, tinterval, rtol, atol, samples, nodes, wt, ewt,
                    fun, a, b, args, plot)

            elif a == -np.Inf and b == np.Inf:
                if nint > 2:
                    # Analytical transformation suggested by K.L. Metlov:
                    alpha = np.tanh(np.arcsinh(2.*interval[1:-1])/2.)
                    tinterval = np.hstack([-1., alpha, 1.])
                    tinterval = _qva_split(tinterval)
                else:
                    tinterval = np.linspace(-1, 1, 11)

                Ifx, errbnd, OK, fcall, vfcall = _qva_Vadapt(
                    _qva_f4, tinterval, rtol, atol, samples, nodes, wt, ewt,
                    fun, a, b, args, plot)

        else:

            tinterval = np.linspace(-1, 1, 11)
            Ifx, errbnd, OK, fcall, vfcall = _qva_Vadapt(
                _qva_f5, tinterval, rtol, atol, samples, nodes, wt, ewt,
                fun, a, b, args, plot)

        if len(Ifx) == 1:
            Ifx = Ifx[0]
            errbnd = errbnd[0]

        if OK:
            status = 0  # success   
        else:
            print('***Integral does not satisfy error test.')
            print('***Approximate bound on error is', errbnd)
            status = 1

        if verbose:
            print(f"Integral: {Ifx} +/- {errbnd}")
            print(f"Function calls: {vfcall}, Function evaluations: {fcall}")

        self.integ = Ifx
        self.errbnd = errbnd
        self.status = status

################################################################################

def quad1d_test1(x, a=3, b=5):
    """
    Gladwell's problem no1. limits x=[0,8]
    Precise result: 0.33333333332074955152

    """
    return np.exp(-a*x)-np.cos(b*np.pi*x)

################################################################################

def quad1d_test2(x):
    """
    Gladwell's problem no2. limits x=[-1,2]
    Precise result: 5.9630898453302550932

    """
    return np.abs(x - 1/np.sqrt(3)) + np.abs(x + 1/np.sqrt(2))

################################################################################

def quad1d_test3(x, a=2.0/3):
    """
    Gladwell's problem no3. limits x=[0,1]
    Precise result: 3 with x^(-2d/3d)

    """
    return x**a

################################################################################

def quad1d_test4(x):
    """
    Gladwell's problem no3. limits x=[0,1]
    Precise result: 2.5066282746310005024

    """
    return np.exp(-x**2/2)

################################################################################

def quad1d_examples():

    print('\ntest 1 ###############################\n')
    ifx = quad1d(quad1d_test1, [0, 8], args=(3, 5), epsrel=0, epsabs=1e-12, singular=1, verbose=1).integ
    print('actual error:', ifx - 0.33333333332074955152)
    print('\ntest 2 ###############################\n')
    ifx = quad1d(quad1d_test2, [-1, 2], epsrel=0, epsabs=1e-12, singular=1, verbose=1).integ
    print('actual error:', ifx - 5.9630898453302550932)
    print('\ntest 2 with breakpoints ##############\n')
    ifx = quad1d(quad1d_test2, [-1, -1/np.sqrt(2), 1/np.sqrt(3), 2], epsrel=0, singular=1, epsabs=1e-12, verbose=1).integ
    print('actual error:', ifx - 5.9630898453302550932)
    print('\ntest 3 ###############################\n')
    ifx = quad1d(quad1d_test3, [0, 1], args=(-2.0/3,), epsrel=0, epsabs=1e-12, singular=1, verbose=1).integ
    print('actual error:', ifx - 3)
    print('\ntest 4 ###############################\n')
    ifx = quad1d(quad1d_test4, [-np.Inf, np.Inf], epsrel=1e-8, epsabs=0, singular=1, verbose=1).integ
    print('actual error:', ifx - 2.5066282746310005024, "\n")
    ifx = quad1d(quad1d_test4, [-8, 8], epsrel=1e-8, epsabs=0, singular=0, verbose=1, plot=1).integ
    print('actual error:', ifx - 2.5066282746310005024)

################################################################################

if __name__ == '__main__':
    quad1d_examples()



"""
#############################################################################

Copyright (C) 2010-2018, Michele Cappellari
E-mail: cappellari_at_astro.ox.ac.uk

Updated versions of the software are available from my web page
http://purl.org/cappellari/software

If you have found this software useful for your research,
I would appreciate an acknowledgement to the use of the
"JAM modelling package of Cappellari (2008)"

    https://ui.adsabs.harvard.edu/abs/2008MNRAS.390...71C

or an explicit reference to the equation given below.

This software is provided as is without any warranty whatsoever.
Permission to use, for non-commercial purposes is granted.
Permission to modify for personal or internal use is granted,
provided this copyright and disclaimer are included unchanged
at the beginning of the file. All other rights are reserved.
In particular, redistribution of the code is not allowed.

#############################################################################

MODIFICATION HISTORY:
    V1.0.0: Written and tested. Michele Cappellari, Oxford, 24 April 2010
    V1.0.1: Use major axis fluxes as reference for isophotes.
        MC, Oxford, 27 July 2011
    V2.0.0: Translated from IDL into Python. MC, Oxford, 21 February 2014
    V2.1.0: Major speedup using histogram. Updated documentation.
        MC, Oxford, 11 March 2016
    V2.2.0: Scale image based on circularized reff.
      - Interpolate reff_maj for increased accuracy.
      - Removed now-unnecessary `npix` keyword.
        MC, Oxford, 22 March 2017
    V2.2.1: Broadcast image creation. MC, Oxford, 22 January 2018

"""

import numpy as np

##############################################################################

def mge_half_light_radius(surf, sigma, q_obs, distance=1e-5):
    """
    Computes the circularized projected half-light radius of an
    MGE model using the approach described after equation (11)
    in Cappellari et al. (2013, MNRAS, 432, 1709)
    https://ui.adsabs.harvard.edu/abs/2013MNRAS.432.1709C

    """
    assert surf.size == sigma.size == q_obs.size, "The MGE components do not match"

    pc = distance*np.pi/0.648         # Constant factor to convert arcsec --> pc
    sigma = sigma*np.sqrt(q_obs)      # Circularize while conserving luminosity
    lum = 2*np.pi*surf*(sigma*pc)**2  # Keep peak surface brightness of each Gaussian
    lum_tot = np.sum(lum)             # If distance is given, lum_tot is in Lsun

    nrad = 50
    rad = np.geomspace(np.min(sigma), np.max(sigma),  nrad)  # arcsec
    lum_r = (lum*(1 - np.exp(-0.5*(rad[:, None]/sigma)**2))).sum(1)
    reff = np.interp(lum_tot/2, lum_r, rad)

    return reff, lum_tot

##############################################################################

def mge_half_light_isophote(surf, sigma, q_obs, distance=1e-5):
    """
    PURPOSE
    -------

        Computes the half-light radius, the  major axis and ellipticity of
        the MGE isophote containing 1/2 of the analytic MGE total light.

        This procedure implements the steps (i)-(iv) described above
        equation (12) in `Cappellari et al. (2013, MNRAS, 432, 1709)
        <https://ui.adsabs.harvard.edu/abs/2013MNRAS.432.1709C>`_

    CALLING SEQUENCE
    ----------------

    .. code-block:: python

        from jampy.mge_half_light_isophote import mge_half_light_isophote

        reff, reff_maj, eps_e, lum_tot = mge_half_light_isophote(
                surf, sigma, q_obs, distance=distance)

    INPUT PARAMETERS
    ----------------

    surf:
        Peak surface brightness of each Gaussian in Lsun/pc**2
    sigma:
        Gaussian dispersion in arcsec
    q_obs: 
        Observed axial ratio of each Gaussian

    OPTIONAL KEYWORD
    ----------------

    distance:
        Galaxy distance in Mpc. This is only required to obtain
        the total luminosity of the MGE model in proper units.

        If the distance is not given, 10pc is assumed. In this case the
        following expression gives the galaxy apparent total magnitude:
        ``mag = sunMag - 2.5*np.log10(lum_tot)``

    OUTPUTS
    -------

    reff:
        Effective (projected half-light) radius Re. This is the
        "circularized" Re=sqrt(Area/pi), where the area is the one
        of the isophote containing 1/2 of the analytic total MGE light.
        This is in the same units as SIGMA (typically arcsec).
    reff_maj:
        Major axis (largest size) of the isophote containing
        1/2 of the total MGE light.
    eps_e:
        Ellipticity of the isophote containing 1/2 of the total MGE
        light, computed from the inertia tensor within the isophote.
    lum_tot:
        Analytic total luminosity in solar luminosities if the
        optional distance in Mpc is provided.

    """
    assert surf.size == sigma.size == q_obs.size, "The MGE components do not match"

    # Scale image based on the circularized Reff and an extreme axial ratio
    reff1, lum_tot = mge_half_light_radius(surf, sigma, q_obs, distance=distance)
    rmax = reff1/np.sqrt(np.min(q_obs).clip(0.1))

    # Create image from MGE model. Only compute one quadrant for symmetry
    npix = 100                   # This gives better than 0.1% accuracy
    scale = rmax/(npix - 0.5)    # image scale in arcsec/pix
    x = np.linspace(scale/2, rmax, npix)  # open interval for bi-symmetry
    xx2, yy2 = np.meshgrid(x**2, x**2)
    image = surf*np.exp(-0.5/sigma**2*(xx2[..., None] + yy2[..., None]/q_obs**2))
    image = image.sum(2)

    # Find enclosed light inside isophotes defined by the MGE flux for the
    # pixels on the major axis, then interpolate to find half-light isophote
    h = np.histogram(image, bins=image[0, ::-1], weights=image)[0]
    lum_r = np.cumsum(h[::-1])
    pc = distance*np.pi/0.648  # Constant factor to convert arcsec --> pc
    surf_e = np.interp(lum_tot/8/(pc*scale)**2, lum_r, image[0, 1:])

    mask = image >= surf_e                  # Pixels inside half-light isophote
    reff = np.sqrt(4*mask.sum()/np.pi)*scale    # Radius of same-area circle
    reff_maj = np.interp(surf_e, image[0, ::-1], x[::-1])

    # Compute coefficients of the diagonal moment of inertia tensor
    # using equation (12) of Cappellari et al. (2013, MNRAS, 432, 1709)
    x2 = image[mask] @ xx2[mask]
    y2 = image[mask] @ yy2[mask]
    eps_e = 1 - np.sqrt(y2/x2)

    return reff, reff_maj, eps_e, lum_tot

##############################################################################



#!/usr/bin/env python

"""
Computes the circular velocity for the MGE approximation of the
axisymmetric Satoh (1980) model given in Table 1 of Cappellari (2020)
https://ui.adsabs.harvard.edu/abs/2020MNRAS.494.4819C
and compares it with the analytic result.

V1.0.0: Michele Cappellari, Oxford, 27 July 2020

"""

from os import path
import numpy as np
import matplotlib.pyplot as plt

import jampy as jampy_package
from jampy.mge_vcirc import mge_vcirc

##############################################################################
def mge_vcirc_example():
    """
    Usage example for mge_vcirc()
    It takes a fraction of a second on a 2GHz computer
    
    """
    inc = 60.  # Inclination in degrees
    distance = 0.648/np.pi  # Mpc (adopt distance where 1" = 1pc)

    # Compute a projected MGE from the intrinsic one in Table 1 of Cappellari (2020)
    jam_dir = path.dirname(path.realpath(jampy_package.__file__))
    filename = path.join(jam_dir, 'examples', 'cappellari2020_table1.txt')
    lg_dens, lg_sigma, qintr = np.loadtxt(filename).T
    dens, sigma, inc_rad = 10**lg_dens, 10**lg_sigma, np.radians(inc)
    rad = np.geomspace(np.min(sigma)/10, np.max(sigma)*3, 30)  # Radii in arcsec where Vcirc must be computed
    
    qobs = np.sqrt((qintr*np.sin(inc_rad))**2 + np.cos(inc_rad)**2)  # Eq.(35) Cappellari (2020)
    surf = np.sqrt(2*np.pi)*dens*sigma*qintr/qobs                    # Eq.(38) Cappellari (2020)

    G = 0.004301    # (km/s)^2 pc/Msun [6.674e-11 SI units (CODATA2018)]
    mbh = 0.01      # Assume a BH 1% of the galaxy mass M=1
    vtrue = np.sqrt(G*rad**2/(3 + rad**2)**1.5 + G*mbh/rad)  # Analytic V_c for M = a = b = 1
    vcirc = mge_vcirc(surf, sigma, qobs, inc, mbh, distance, rad)

    plt.plot(np.log10(rad), vtrue, '-', label="Analytic")
    plt.plot(np.log10(rad), vcirc, 'o', label="MGE approximation")
    plt.xlabel('lg R (arcsec)')
    plt.ylabel(r'$V_{circ}$ (km/s)')
    plt.title(r"Circular velocity of Satoh model with $M=1$ and $M_{\rm BH}=M/100$")
    plt.legend()

##############################################################################

if __name__ == '__main__':
    
    plt.clf()
    mge_vcirc_example()
    plt.pause(1)



"""
    Copyright (C) 2019-2024, Michele Cappellari

    E-mail: michele.cappellari_at_physics.ox.ac.uk

    Updated versions of the software are available from my web page
    https://purl.org/cappellari/software

    This software is provided as is without any warranty whatsoever.
    Permission to use, for non-commercial purposes is granted.
    Permission to modify for personal or internal use is granted,
    provided this copyright and disclaimer are included unchanged
    at the beginning of the file. All other rights are reserved.
    In particular, redistribution of the code is not allowed.

Changelog
---------

V1.0.0: Michele Cappellari, Oxford, 08 November 2019
    - Written and tested.
Vx.x.x: Additional changes are documented in the CHANGELOG of the JamPy package.

"""
import numpy as np
import matplotlib.pyplot as plt
from scipy import ndimage, special
from time import perf_counter as clock

from plotbin.plot_velfield import plot_velfield
from jampy.quad1d import quad1d
from jampy.quad2d import quad2d

##############################################################################

def bilinear_interpolate(xv, yv, im, xout, yout):
    """
    The input array has size im[ny,nx] as in the output
    of im = f(meshgrid(xv, yv))
    xv and yv are vectors of size nx and ny respectively.

    """
    ny, nx = im.shape
    assert (nx, ny) == (xv.size, yv.size), "Input arrays dimensions do not match"

    xi = (nx - 1.)/(xv[-1] - xv[0]) * (xout - xv[0])
    yi = (ny - 1.)/(yv[-1] - yv[0]) * (yout - yv[0])

    return ndimage.map_coordinates(im.T, [xi, yi], order=1, mode='nearest')

##############################################################################

def density(R, z, dens, sigma, qintr):
    """ Density for each luminous Gaussian at (R, z) """

    nu = dens*np.exp(-0.5/sigma**2*(R**2 + (z/qintr)**2))  # Cappellari (2008) eq.(13)

    return nu

##############################################################################

def intrinsic_moments_cyl_bh(R, z, dens_lum, sigma_lum, qintr_lum, mbh, beta, gamma, all_mom):
    """ Compute analytic intrinsic moments for the black hole alone """

    G = 0.004301    # (km/s)^2 pc/Msun [6.674e-11 SI units (CODATA2018)]
    h = sigma_lum*qintr_lum*np.sqrt(2)
    r = np.sqrt(R**2 + z**2)
    nu = density(R, z, dens_lum, sigma_lum, qintr_lum)
    nu /= nu.sum()
    bani = 1/(1 - beta)  # Anisotropy ratio b = (sig_R/sig_z)**2

    sig2z = G*mbh*nu*(1/r - special.erfcx(r/h)*np.sqrt(np.pi)/h)
    v2phi = G*mbh*R**2/r**3*(1 - bani)*nu + (
            1 + (1 - qintr_lum**2)*(R/(sigma_lum*qintr_lum))**2)*bani*sig2z
    arr = [sig2z, v2phi]

    if all_mom:
        sig2R = bani*sig2z
        sig2phi = (1 - gamma)*sig2R
        arr.extend([sig2R, sig2phi])

    return np.sum(arr, 1)

##############################################################################

def integrand_cyl(u, R, z,
                  dens_lum, sigma_lum, qintr_lum,
                  dens_pot, sigma_pot, qintr_pot,
                  beta, gamma, logistic, all_mom):
    """
    Compute all the non-zero JAM first and second intrinsic velocity moments
    using the formulas from Cappellari (2008, MNRAS, 390, 71; hereafter C08).
    https://ui.adsabs.harvard.edu/abs/2008MNRAS.390...71C

    """
    dens_lum = dens_lum[:, None, None]
    sigma_lum = sigma_lum[:, None, None]
    qintr_lum = qintr_lum[:, None, None]

    dens_pot = dens_pot[None, :, None]
    sigma_pot = sigma_pot[None, :, None]
    qintr_pot = qintr_pot[None,:, None]

    # DE Change of variables for Chandrasekhar u-integral (Sec.6.2 of Cappellari 2020)
    # np.arcsinh(np.log([rmin, rmax])*2/np.pi) -> [0, inf]
    x = np.exp(np.sinh(u)*np.pi/2)
    duds = x*np.cosh(u)*np.pi/2
    u = x[None, None, :]

    if logistic:  # Variable anisotropy, same for all Gaussians
        zab, beta0, betainf, alpha = beta
        beta = (beta0 + betainf*(abs(z)/zab)**alpha)/(1 + (abs(z)/zab)**alpha)
        bani = 1/(1 - beta)  # Anisotropy ratio b = (sig_R/sig_z)**2
    else:  # Constant anisotropy, different per Gaussian
        bani = 1/(1 - beta[:, None, None]) if np.ptp(beta) else 1/(1 - beta[0])

    # Tracer Gaussians
    R2, z2 = R**2, z**2
    a = -1/(2*sigma_lum**2)
    b = a/qintr_lum**2
    exp1 = dens_lum*np.exp(a*R2 + b*z2)

    # Mass Gaussians
    G = 0.004301    # (km/s)^2 pc/Msun [6.674e-11 SI units (CODATA2018)]
    u1, qu = 1 + u, qintr_pot**2 + u
    tmp = -1/(2*sigma_pot**2)
    c, d = tmp/u1, tmp/qu
    cst = 2*np.pi*G*dens_pot*qintr_pot*sigma_pot**2
    exp2 = cst*np.exp(c*R2 + d*z2)/(u1*np.sqrt(qu))

    bd = d/(b + d)
    sig2z = exp1*exp2*bd
    v2phi = (bani + 2*R2*((a + c)*bani - c/bd))*sig2z
    arr = [sig2z, v2phi]

    if all_mom:
        gamma = gamma[:, None, None] if np.ptp(gamma) else gamma[0]
        sig2R = bani*sig2z
        sig2phi = (1 - gamma)*sig2R
        arr.extend([sig2R, sig2phi])

    return duds*np.sum(arr, (1, 2))

##############################################################################

def intrinsic_moments_cyl(R, z,
                          dens_lum, sigma_lum, qintr_lum,
                          dens_pot, sigma_pot, qintr_pot,
                          mbh, beta, gamma, logistic, epsrel):
    """ Numerical quadratures of the Jeans solution """

    mds, mxs = np.median(sigma_lum), np.max(sigma_lum)
    xlim = np.arcsinh(np.log([1e-7*mds, 1e3*mxs])*2/np.pi)
    nu = density(R, z, dens_lum, sigma_lum, qintr_lum).sum()

    all_mom = True if (not logistic) and (np.ptp(beta) or np.ptp(gamma)) else False
    args = [R, z, dens_lum, sigma_lum, qintr_lum, dens_pot, sigma_pot, qintr_pot,
            beta, gamma, logistic, all_mom]
    mom = quad1d(integrand_cyl, xlim, epsrel=epsrel, args=args).integ/nu

    if logistic:
        zab, beta0, betainf, alpha = beta
        beta = (beta0 + betainf*(abs(z)/zab)**alpha)/(1 + (abs(z)/zab)**alpha)
        zag, gamma0, gammainf, alpha = gamma
        gamma = (gamma0 + gammainf*(abs(z)/zag)**alpha)/(1 + (abs(z)/zag)**alpha)
    else:
        if not np.ptp(beta): beta = beta[0]
        if not np.ptp(gamma): gamma = gamma[0]

    if mbh > 0:  
        mom += intrinsic_moments_cyl_bh(
            R, z, dens_lum, sigma_lum, qintr_lum, mbh, beta, gamma, all_mom)

    if all_mom:
        sig2z, v2phi, sig2R, sig2phi = mom
    else:
        sig2z, v2phi = mom
        sig2R = sig2z/(1 - beta)
        sig2phi = sig2R*(1 - gamma)

    return sig2R, sig2z, sig2phi, v2phi, nu

##############################################################################

def integrand_sph_bh(t, r, th, dens_lum, sigma_lum, qintr_lum, beta, gamma, logistic, mbh, all_mom):
    """ Solution of the spherically-aligned Jeans equations for a black hole alone """

    # TANH Change of variables for Jeans r-integral (Sec.6.2 of Cappellari 2020)
    # np.log([rmin, rmax]) -> [r, inf]
    drdt = np.exp(t)
    r1 = r + drdt[:, None]

    if logistic:    # Variable anisotropy, same for all Gaussians
        rab, beta0, betainf, alpha = beta
        fun = (r1/r)**(2*beta0)
        fun *= ((1 + (r1/rab)**alpha)/(1 + (r/rab)**alpha))**(2*(betainf - beta0)/alpha)
        beta = beta0 + (betainf - beta0)/(1 + (rab/r)**alpha)
    else:           # Constant anisotropy, different per Gaussian
        fun = (r1/r)**(2*beta)

    # Tracer Gaussians
    s2_lum = sigma_lum**2
    q2_lum = qintr_lum**2
    rs = fun*(r*np.sin(th))**2
    aa = -r1**2/(2*q2_lum*s2_lum)
    bb = (1 - q2_lum)/(2*q2_lum*s2_lum)*rs
    ex1 = dens_lum*np.exp(aa + bb)

    G = 0.004301    # (km/s)^2 pc/Msun [6.674e-11 SI units (CODATA2018)]
    sig2r = G*fun*ex1*mbh/r1**2
    v2phi = (1 - beta)*(1 + 2*bb)*sig2r
    arr = [sig2r, v2phi]

    if all_mom:
        sig2th = (1 - beta)*sig2r       # Anisotropy beta = 1 - (sig_th/sig_r)**2
        sig2phi = (1 - gamma)*sig2r     # Anisotropy gamma = 1 - (sig_phi/sig_r)**2
        arr.extend([sig2th, sig2phi])

    return drdt*np.sum(arr, 2)

##############################################################################

def integrand_tan_dth_pot(u, R2, z2, dens_pot, sigma_pot, qintr_pot):
    """
    Returns the integrand of the tan(th)*D[pot, th] derivative
    of the MGE potential at (r, th).
    This is equation (51) of Cappellari (2020, MNRAS, 494, 4819)
    https://ui.adsabs.harvard.edu/abs/2020MNRAS.494.4819C

    """
    # DE Change of variables for Chandrasekhar u-integral (Sec.6.2 of Cappellari 2020)
    # np.arcsinh(np.log([rmin, rmax])*2/np.pi) -> [0, inf]
    x = np.exp(np.sinh(u)*np.pi/2)
    duds = x*np.cosh(u)*np.pi/2
    u = x[:, None]

    G = 0.004301    # (km/s)^2 pc/Msun [6.674e-11 SI units (CODATA2018)]
    u1, qu = 1 + u, qintr_pot**2 + u
    d = 2*np.pi*G*dens_pot*qintr_pot*(qintr_pot**2 - 1)*R2
    e = np.exp(-0.5/sigma_pot**2*(R2/u1 + z2/qu))
    tan_dth_pot = d*e/(u1**2*qu**1.5)

    return duds*tan_dth_pot.sum(1)   # u.size

##############################################################################

def integrand_sph(s, t, r, th, dens_lum, sigma_lum, qintr_lum, beta, gamma,
                  logistic, dens_pot, sigma_pot, qintr_pot, all_mom):
    """
    Solution of the spherically-aligned Jeans equations for an MGE
    from eq.(52)-(54) of Cappellari (2020, MNRAS, 494, 4819)
    https://ui.adsabs.harvard.edu/abs/2020MNRAS.494.4819C

    """
    dens_lum = dens_lum[:, None, None]
    s2_lum = (sigma_lum**2)[:, None, None]
    q2_lum = (qintr_lum**2)[:, None, None]

    dens_pot = dens_pot[None, :, None]
    s2_pot = (sigma_pot**2)[None, :, None]
    qintr_pot = qintr_pot[None, :, None]

    # DE Change of variables for Chandrasekhar u-integral (Sec.6.2 of Cappellari 2020)
    # np.arcsinh(np.log([rmin, rmax])*2/np.pi) -> [0, inf]
    x = np.exp(np.sinh(s)*np.pi/2)
    duds = x*np.cosh(s)*np.pi/2
    u = x[None, None, :]

    # TANH Change of variables for Jeans r-integral (Sec.6.2 of Cappellari 2020)
    # np.log([rmin, rmax]) -> [r, inf]
    drdt = np.exp(t)
    r1 = r + drdt[None, None, :]

    if logistic:        # Variable anisotropy, same for all Gaussians
        rab, beta0, betainf, alpha = beta
        fun = (r1/r)**(2*beta0)
        fun *= ((1 + (r1/rab)**alpha)/(1 + (r/rab)**alpha))**(2*(betainf - beta0)/alpha)
        beta = beta0 + (betainf - beta0)/(1 + (rab/r)**alpha)
    else:               # Constant anisotropy, different per Gaussian
        beta = beta[:, None, None] if np.ptp(beta) else beta[0]
        fun = (r1/r)**(2*beta)

    # Tracer Gaussians
    rs = fun*(r*np.sin(th))**2
    aa = -r1**2/(2*q2_lum*s2_lum)
    bb = (1 - q2_lum)/(2*q2_lum*s2_lum)*rs
    ex1 = dens_lum*np.exp(aa + bb)

    # Mass Gaussians
    G = 0.004301    # (km/s)^2 pc/Msun [6.674e-11 SI units (CODATA2018)]
    u1, qu = 1 + u, qintr_pot**2 + u
    cc = -r1**2/(2*qu*s2_pot)
    dd = (u1 - qu)/(2*qu*s2_pot*u1)*rs
    cst = 2*np.pi*G*dens_pot*qintr_pot*r1
    ex2 = cst*np.exp(cc + dd)/(u1*qu**1.5)

    sig2r = fun*ex1*ex2
    v2phi = (1 - beta)*(1 + 2*(bb + dd))*sig2r
    arr = [sig2r, v2phi]

    if all_mom:
        gamma = gamma[:, None, None] if np.ptp(gamma) else gamma[0]
        sig2th = (1 - beta)*sig2r       # Anisotropy beta = 1 - (sig_th/sig_r)**2
        sig2phi = (1 - gamma)*sig2r     # Anisotropy gamma = 1 - (sig_phi/sig_r)**2
        arr.extend([sig2th, sig2phi])

    return duds*drdt*np.sum(arr, (1, 2))

##############################################################################

def intrinsic_moments_sph(R, z,
                          dens_lum, sigma_lum, qintr_lum,
                          dens_pot, sigma_pot, qintr_pot,
                          mbh, beta, gamma, logistic, epsrel):
    """ Numerical quadratures of the Jeans solution """

    r = np.sqrt(R**2 + z**2)
    th = np.arctan2(R, z)  # Angle from symmetry axis z

    mds, mxs = np.median(sigma_lum), np.max(sigma_lum)
    xlim = np.arcsinh(np.log([1e-7*mds, 1e3*mxs])*2/np.pi)
    ylim = np.log([1e-6*mds, 3*mxs])
    args = [r, th, dens_lum, sigma_lum, qintr_lum, beta, gamma, logistic]
    nu = density(R, z, dens_lum, sigma_lum, qintr_lum).sum()

    all_mom = True if (not logistic) and (np.ptp(beta) or np.ptp(gamma)) else False
    args_mge = args + [dens_pot, sigma_pot, qintr_pot, all_mom]
    mom = quad2d(integrand_sph, xlim, ylim, epsrel=epsrel, args=args_mge).integ
    if mbh > 0:
        mom += quad1d(integrand_sph_bh, ylim, epsrel=epsrel, args=args+[mbh, all_mom]).integ

    if all_mom:
        sig2r, v2phi, sig2th, sig2phi = mom/nu
    else:
        sig2r, v2phi = mom/nu
        if logistic:
            rab, beta0, betainf, alpha = beta
            beta = beta0 + (betainf - beta0)/(1 + (rab/r)**alpha)
            rag, gamma0, gammainf, alpha = gamma
            gamma = gamma0 + (gammainf - gamma0)/(1 + (rag/r)**alpha)
        else:
            beta = beta[0]
            gamma = gamma[0]
        sig2th = sig2r*(1 - beta)
        sig2phi = sig2r*(1 - gamma)

    v2phi += quad1d(integrand_tan_dth_pot, xlim, epsrel=epsrel,
                    args=(R**2, z**2, dens_pot, sigma_pot, qintr_pot)).integ

    return sig2r, sig2th, sig2phi, v2phi, nu

##############################################################################

def intrinsic_moments(Rbin, zbin, dens_lum, sigma_lum, qintr_lum, dens_pot,
                      sigma_pot, qintr_pot, mbh, beta, gamma, logistic, nrad,
                      nang, epsrel, interp, proj_cyl, align):

    fun = intrinsic_moments_cyl if align == 'cyl' else intrinsic_moments_sph

    if nrad*nang > Rbin.size or (not interp):  # Just calculate values
        mom = [fun(Rj, zj, dens_lum, sigma_lum, qintr_lum, dens_pot,
                   sigma_pot, qintr_pot, mbh, beta, gamma, logistic, epsrel)
               for (Rj, zj) in zip(Rbin, zbin)]
        sig2r, sig2th, sig2phi, v2phi, nu = np.array(mom).T
    else:
        irp = mom_interp(Rbin, zbin, dens_lum, sigma_lum, qintr_lum, dens_pot,
                         sigma_pot, qintr_pot, mbh, beta, gamma, logistic, nrad,
                         nang, epsrel=epsrel, align=align)
        sig2r, sig2th, sig2phi, v2phi, nu = irp.get_moments(Rbin, zbin)

    # Project the velocity dispersion tensor to cylindrical coordinates
    if proj_cyl and (align == 'sph'):
        th = np.arctan2(Rbin, zbin)  # Angle from symmetry axis z
        sig2R = sig2th*np.cos(th)**2 + sig2r*np.sin(th)**2
        sig2z = sig2th*np.sin(th)**2 + sig2r*np.cos(th)**2
        sig2r, sig2th = sig2R, sig2z

    return sig2r, sig2th, sig2phi, v2phi, nu

##############################################################################

class mom_interp:

    def __init__(self, xbin, ybin,
                 dens_lum, sigma_lum, qintr_lum,
                 dens_pot, sigma_pot, qintr_pot,
                 mbh, beta, gamma, logistic, nrad, nang, epsrel=1e-2,
                 rmin=None, rmax=None, align='cyl'):
        """ Initializes model values on a grid for subsequent interpolation """

        fun = intrinsic_moments_cyl if align == 'cyl' else intrinsic_moments_sph

        # Define parameters of polar grid for interpolation
        w = sigma_lum < np.max(np.abs(xbin))  # Characteristic MGE axial ratio in observed range
        self.qmed = np.median(qintr_lum) if w.sum() < 3 else np.median(qintr_lum[w])

        if rmin is None or rmax is None:
            rell2 = xbin**2 + (ybin/self.qmed)**2

        self.rmin = np.sqrt(np.min(rell2)) if rmin is None else rmin
        self.rmax = np.sqrt(np.max(rell2)) if rmax is None else rmax

        # Make linear grid in log of elliptical radius RAD and eccentric anomaly ECC
        rad =  np.geomspace(self.rmin, self.rmax, nrad)
        self.logRad = np.log(rad)
        self.ang = np.linspace(0, np.pi/2, nang)
        rellGrid, eccGrid = map(np.ravel, np.meshgrid(rad, self.ang))
        R = rellGrid*np.cos(eccGrid)
        z = rellGrid*np.sin(eccGrid)*self.qmed  # ecc=0 on equatorial plane

        mom = [fun(Rj, zj, dens_lum, sigma_lum, qintr_lum, dens_pot,
                   sigma_pot, qintr_pot, mbh, beta, gamma, logistic, epsrel)
               for (Rj, zj) in zip(R, z)]

        self.sig2r, self.sig2th, self.sig2phi, self.v2phi, nu = np.array(mom).T.reshape((5, nang, nrad))
        self.dens_lum = dens_lum
        self.sigma_lum = sigma_lum
        self.qintr_lum = qintr_lum

##############################################################################

    def get_moments(self, R, z):
        """
        Fast linearly-interpolated model values at the set
        of (R, z) coordinates from pre-computed values.
        Interpolation of non-weighted kinematics for accuracy.
        The returned density is analytic, not interpolated.

        """
        r1 = 0.5*np.log((R**2 + (z/self.qmed)**2).clip(self.rmin**2))  # Log elliptical radius of input (R,z)
        e1 = np.arctan2(np.abs(z/self.qmed), np.abs(R))  # Eccentric anomaly of input (R,z)
        bint = lambda x: bilinear_interpolate(self.logRad, self.ang, x, r1, e1)
        sig2r, sig2th, sig2phi, v2phi = map(bint, [self.sig2r, self.sig2th, self.sig2phi, self.v2phi])
        nu = density(R[:, None], z[:, None], self.dens_lum, self.sigma_lum, self.qintr_lum).sum(-1)

        return sig2r, sig2th, sig2phi, v2phi, nu

##############################################################################

class jam_axi_intr:
    """
    jam_axi_intr
    ============

    Purpose
    -------

    This procedure calculates all the intrinsic first and second velocity
    moments for an anisotropic axisymmetric galaxy model.

    This program is useful e.g. to model the kinematics of galaxies
    like our Milky Way, for which the intrinsic moments can be observed
    directly, or to compute starting conditions for N-body numerical
    simulations of galaxies.

    Two assumptions for the orientation of the velocity ellipsoid are supported:

    - The cylindrically-aligned ``(R, z, phi)`` solution was presented in
      `Cappellari (2008) <https://ui.adsabs.harvard.edu/abs/2008MNRAS.390...71C>`_

    - The spherically-aligned ``(r, th, phi)`` solution was presented in
      `Cappellari (2020) <https://ui.adsabs.harvard.edu/abs/2020MNRAS.494.4819C>`_

    Calling Sequence
    ----------------

    .. code-block:: python

        from jampy.jam_axi_intr import jam_axi_intr

        jam = jam_axi_intr(
                 dens_lum, sigma_lum, qintr_lum, dens_pot, sigma_pot, qintr_pot,
                 mbh, Rbin, zbin, align='cyl', beta=None, data=None, epsrel=1e-2,
                 errors=None, gamma=None, goodbins=None, interp=True,
                 logistic=False, ml=None, nang=10, nodots=False, nrad=20,
                 plot=True, proj_cyl=False, quiet=False)

        # The meaning of the output is different depending on `align`
        sig2R, sig2z, sig2phi, v2phi = jam.model  # with align='cyl'
        sig2r, sig2th, sig2phi, v2phi = jam.model  # with align='sph'

        jam.plot()   # Generate data/model comparison

    Input Parameters
    ----------------

    dens_lum: array_like with shape (n,)
        vector containing the peak value of the MGE Gaussians describing
        the intrinsic density of the tracer population for which the kinematics
        is derived.
        The units are arbitrary as they cancel out in the final results.
        Typical units are e.g. ``Lsun/pc^3`` (solar luminosities per ``parsec^3``)
    sigma_lum: array_like with shape (n,)
        vector containing the dispersion (sigma) in ``pc`` of the MGE
        Gaussians describing the galaxy kinematic-tracer population.
    qintr_lum: array_like with shape (n,)
        vector containing the intrinsic axial ratio (q) of the MGE
        Gaussians describing the galaxy kinematic-tracer population.
    surf_pot: array_like with shape (m,)
        vector containing the peak value of the MGE Gaussians
        describing the galaxy total-mass density in units of ``Msun/pc^3``
        (solar masses per ``parsec^3``). This is the MGE model from which the
        model gravitational potential is computed.
    sigma_pot: array_like with shape (m,)
        vector containing the dispersion in ``pc`` of the MGE
        Gaussians describing the galaxy total-mass density.
    qintr_pot: array_like with shape (m,)
        vector containing the intrinsic axial ratio of the MGE
        Gaussians describing the galaxy total-mass density.
    mbh: float
        Mass of a nuclear supermassive black hole in solar masses.
    Rbin: array_like with shape (p,)
        Vector with the ``R`` coordinates in ``pc`` of the bins (or pixels) at
        which one wants to compute the model predictions. This is the first
        cylindrical coordinate ``(R, z)`` with the galaxy center at ``(0,0)``.

        There is a singularity at ``(0, 0)`` which should be avoided by the user
        in the input coordinates.
    zbin: array_like with shape (p,)
        Vector with the ``z`` coordinates in ``pc`` of the bins (or pixels) at
        which one wants to compute the model predictions. This is the second
        cylindrical coordinate ``(R, z)``, with the z-axis coincident with the
        galaxy symmetry axis.

    Optional Keywords
    -----------------

    align: {'cyl', 'sph'} optional
        If ``align='cyl'`` the program computes the solution of the Jeans
        equations with cylindrically-aligned velocity ellipsoid, presented
        in `Cappellari (2008)`_. If ``align='sph'`` the spherically-aligned
        solution of `Cappellari (2020)`_ is returned.
    beta: array_like with shape (n,) or (4,)
        Vector with the axial anisotropy of the individual kinematic-tracer
        MGE Gaussians (Default: ``beta=np.zeros(n)``)::

            beta = 1 - (sigma_th/sigma_r)^2  # with align='sph'
            beta = 1 - (sigma_z/sigma_R)^2   # with align='cyl'

        When ``logistic=True`` the procedure assumes::

            beta = [r_a, beta_0, beta_inf, alpha]

        and the anisotropy of the whole JAM model varies as a logistic
        function of the logarithmic spherical radius (with ``align='sph'``) or
        of the logarithmic distance from the equatorial plane (with ``align='cyl'``)::

            beta(r) = beta_0 + (beta_inf - beta_0)/[1 + (r_a/r)^alpha]    # with align='sph'
            beta(z) = beta_0 + (beta_inf - beta_0)/[1 + (z_a/|z|)^alpha]  # with align='cyl'

        Here ``beta_0`` represents the anisotropy at ``r = 0``, ``beta_inf`` is
        the anisotropy at ``r = inf`` and ``r_a`` is the anisotropy transition
        radius (in arcsec like ``sigma_lum``), with ``alpha`` controlling the
        sharpness of the transition. 
        In the special case ``beta_0 = 0, beta_inf = 1, alpha = 2`` the
        anisotropy variation reduces to the form by Osipkov & Merritt, but the
        extra parameters allow for much more realistic anisotropy profiles. See
        details and an application in `Simon, Cappellari & Hartke (2024)
        <https://ui.adsabs.harvard.edu/abs/2024MNRAS.527.2341S>`_.
    data: array_like of shape (4, p), optional
        Four input vectors with the observed values of:

        - ``[sigR, sigz, sigphi, vrms_phi]`` in ``km/s``, when ``align='cyl'``
          (or ``align='sph'`` and ``proj_cyl=True``).

          ``vrms_phi`` is the square root of the velocity second moment in the
          tangential direction. If the velocities ``vphi_j`` are measured from
          individual stars then ``vrms_phi = sqrt(mean(vphi_j^2))``.
          One can also use the relation ``vrms_phi = sqrt(vphi^2 + sigphi^2)``,
          where ``vphi = mean(vphi_j)`` and ``sigphi = std(vphi_j)``

        - ``[sigr, sigth, sigphi, vrms_phi]`` in ``km/s``, when ``align='sph'``,
          where ``vrms_phi`` is defined above.

    epsrel: float, optional
        Relative error requested for the numerical quadratures, before
        interpolation (Default: ``epsrel=1e-2``).
    errors: array_like of shape (4, p), optional
        ``1sigma`` uncertainties on ``data``, in the same format (default 5 ``km/s``).
    gamma: array_like with shape (n,)
        Vector with the tangential anisotropy of the individual kinematic-tracer
        MGE Gaussians (Default: ``gamma=np.zeros(n)``)::

            gamma = 1 - (sigma_phi/sigma_r)^2  # with align='sph'
            gamma = 1 - (sigma_phi/sigma_R)^2  # with align='cyl'

        When ``logistic=True`` the procedure assumes::

            gamma = [r_a, gamma_0, gamma_inf, alpha]

        and the anisotropy of the whole JAM model varies as a logistic
        function of the logarithmic spherical radius (with ``align='sph'``) or
        of the logarithmic distance from the equatorial plane (with ``align='cyl'``)::

            gamma(r) = gamma_0 + (gamma_inf - gamma_0)/[1 + (r_a/r)^alpha]    # with align='sph'
            gamma(z) = gamma_0 + (gamma_inf - gamma_0)/[1 + (z_a/|z|)^alpha]  # with align='cyl'

        Here ``gamma_0`` represents the anisotropy at ``r = 0``, ``gamma_inf``
        is the anisotropy at ``r = inf`` and ``r_a`` is the anisotropy
        transition radius (in arcsec like ``sigma_lum``), with ``alpha``
        controlling the sharpness of the transition. In the special case
        ``gamma_0 = 0, gamma_inf = 1, alpha = 2`` the anisotropy variation
        reduces to the form by Osipkov & Merritt, but the extra parameters
        allow for much more realistic anisotropy profiles.

        IMPORTANT: ``gamma`` only affects the projected first velocity moments.
        The projected second moments are rigorously independent of ``gamma``.
    goodbins: array_like with shape (4, p), optional
        Boolean vector of the same shape as ``data`` with values ``True``
        for the bins which have to be included in the fit (if requested) and
        ``chi^2`` calculation (Default: fit all bins).
    interp: bool, optional
        If ``interp=False`` no interpolation is performed and the model is
        computed at every set of input (R, z) coordinates.
        If ``interp=True`` (default), the model is interpolated if the number
        of requested input (R, z) coordinates is larger than ``nang*nrad``.
    logistic: bool, optional
        When ``logistic=True``, JAM interprets the anisotropy parameters
        ``beta`` and ``gamma`` as defining a 4-parameters logistic function.
        See the documentation of the anisotropy keywords for details.
        (Default ``logistic=False``)
    ml: float, optional
        Mass-to-light ratio M/L. If ``ml=None`` (default) the M/L is fitted to
        the data and the best-fitting value is returned in output.
        The ``mbh`` is also scaled and becomes ``mbh*ml``.
        If ``ml=1`` no scaling is applied to the model.
    nang: int, optional
        The number of linearly-spaced intervals in the eccentric anomaly at
        which the model is evaluated before interpolation (default: ``nang=10``).
    nodots: bool, optional
        Set to ``True`` to hide the dots indicating the centers of the bins in
        the two-dimensional map (default ``False``).
    nrad: int, optional
        The number of logarithmically spaced radial positions at which the
        model is evaluated before interpolation. One may want to increase this
        value if the model has to be evaluated over many orders of magnitude in
        radius (default: ``nrad=20``).
    plot: bool, optional
        If ``plot=True`` (default) and ``data is not None``, produce a plot of
        the data-model comparison at the end of the calculation.
    proj_cyl: bool, optional
        If ``align='sph'`` and ``proj_cyl=True``, the function projects the
        spherically-aligned moments to cylindrical coordinates and returns the
        ``[sig2R, sig2z, sig2phi, v2phi]`` components as in the case
        ``align='cyl'``. This is useful for a direct comparison of results with
        either the spherical or cylindrical alignment, as it allows one to fit
        the same data with both modelling assumptions.
    quiet: bool, optional
        If ``quiet=False`` (default), print the best-fitting M/L and chi2 at
        the end for the calculation.

    Output Parameters
    -----------------

    Returned as attributes of the ``jam_axi_intr`` class.

    .chi2: float
        Reduced chi^2 (chi^2/DOF) describing the quality of the fit::

            d = (data/errors)[goodbins]
            m = (model/errors)[goodbins]
            chi2 = ((d - m)**2).sum()/goodbins.sum()

    .flux: array_like  with shape (p,)
        Vector with the MGE luminosity density at each ``(R, z)`` location in
        ``Lsun/pc^3``, used to plot the isophotes on the model results.
    .ml: float
        Best fitting M/L. This value is fitted while ignoring ``sigphi`` and it
        is strictly independent of the adopted tangential anisotropy ``gamma``.
    .model: array_like with shape (4, p)
        - Contains ``[sig2R, sig2z, sig2phi, v2phi]`` with ``align='cyl'``

        - Contains ``[sig2r, sig2th, sig2phi, v2phi]`` with ``align='sph'``

        where the above quantities are defined as follows:

        sig2R (sig2r): array_like with shape (p,)
            squared intrinsic dispersion in ``(km/s)^2`` along the R (r)
            direction at each ``(R, z)`` location.

        sig2z (sig2th): array_like with shape (p,)
            squared intrinsic dispersion in ``(km/s)^2`` along the z (th)
            direction at each ``(R, z)`` location.

        sig2phi: array_like with shape (p,)
            squared intrinsic dispersion in ``(km/s)^2``  along the
            tangential ``phi`` direction at each ``(R, z)`` location.

        v2phi: array_like with shape (p,)
            the second velocity moment in ``(km/s)^2`` along the
            tangential ``phi`` direction at each ``(R, z)`` location.

        The mean velocity along the tangential direction can be computed as
        ``vphi = np.sqrt(v2phi - sig2phi)``

        NOTE: I return squared velocities instead of taking the square root,
        to allow for negative values (unphysical solutions).

    ###########################################################################
    """

    def __init__(self, dens_lum, sigma_lum, qintr_lum, dens_pot, sigma_pot,
                 qintr_pot, mbh, Rbin, zbin, align='cyl', beta=None, data=None,
                 epsrel=1e-2, errors=None, gamma=None, goodbins=None,
                 interp=True, logistic=False, ml=None, nang=10, nodots=False,
                 nrad=20, plot=True, proj_cyl=False, quiet=False):

        assert align in ['sph', 'cyl'], "`align` must be 'sph' or 'cyl'"
        assert (ml is None) or (ml > 0), "The input `ml` must be positive"
        if beta is None:
            beta = np.zeros_like(dens_lum)  # Anisotropy parameter beta = 1 - (sig_th/sig_r)**2  (beta=0-->circle)
        if gamma is None:  # Anisotropy parameter beta = 1 - (sig_th/sig_r)**2
            if logistic:
                gamma = [1, 0, 0, 1]    # [r_a, gamma_0, gamma_inf, alpha]
            else:
                gamma = np.zeros_like(beta)
        assert (dens_lum.size == sigma_lum.size == qintr_lum.size) \
               and ((len(beta) == 4 and logistic) or (len(beta) == dens_lum.size)) \
               and (len(beta) == len(gamma)), "The luminous MGE components and anisotropies do not match"
        assert dens_pot.size == sigma_pot.size == qintr_pot.size, "The total mass MGE components do not match"
        assert Rbin.size == zbin.size, "Rbin and zbin do not match"

        if data is not None:
            assert len(data) == 4, "`data` must contain four vectors"
            if errors is None:
                errors = np.full_like(data, 5)  # Constant 5 km/s errors
            if goodbins is None:
                goodbins = np.ones_like(data, dtype=bool)
            else:
                assert goodbins.dtype == bool, "goodbins must be a boolean vector"
                assert np.any(goodbins), "goodbins must contain some True values"
            assert Rbin.size == len(data[0]) == len(errors[0]) == len(goodbins[0]), \
                "(data, errors, goodbins) and (Rbin, zbin) do not match"
        if goodbins is None:
            goodbins = np.ones_like(data, dtype=bool)
        else:
            assert goodbins.dtype == bool, "goodbins must be a boolean vector"
            assert np.any(goodbins), "goodbins must contain some True values"

        t = clock()

        # model contains [sig2r, sig2th, sig2phi, v2phi]
        *model, nu = intrinsic_moments(Rbin, zbin, dens_lum, sigma_lum,
            qintr_lum, dens_pot, sigma_pot, qintr_pot, mbh, beta, gamma,
            logistic, nrad, nang, epsrel, interp, proj_cyl, align)

        if not quiet:
            if nrad*nang > Rbin.size or (not interp):
                print("Just computed values")
            else:
                print("Performed interpolation")
            print(f'jam_axi_intr_{align} elapsed time (sec): {clock() - t:.2f}')

        if data is None:
            chi2 = None
            if ml is None:
                ml = 1.
        else:
            model0 = np.sqrt(np.clip(model, 0, None))   # sqrt([sig2r, sig2th, sig2phi, v2phi])

            # Exclude sig2phi from M/L calculation
            if (ml is None) or (ml <= 0):
                m = np.delete(model0, 2, 0).ravel()     # sqrt([sig2r, sig2th, v2phi])
                d = np.delete(data, 2, 0).ravel()
                e = np.delete(errors, 2, 0).ravel()
                ok = np.delete(goodbins, 2, 0).ravel()
                d, m = (d/e)[ok], (m/e)[ok]
                ml = ((d @ m)/(m @ m))**2   # eq. (51) of Cappellari (2008, MNRAS)

            # Use full data for chi2 calculation
            model0 *= np.sqrt(ml)
            d, m, e, ok = map(np.ravel, [data, model0, errors, goodbins])
            chi2 = (((d[ok] - m[ok])/e[ok])**2).sum()/ok.sum()

        self.Rbin = Rbin
        self.zbin = zbin
        self.align = align
        self.model = np.array(model)*ml  # model ~ V^2
        self.data = data
        self.errors = errors
        self.flux = nu
        self.ml = ml
        self.chi2 = chi2
        self.goodbins = goodbins

        if plot and (data is not None):
            self.plot(nodots)

##############################################################################

    def plot(self, nodots=False):
        """ Data-model comparison for jam_axi_intr """

        plt.figure('jam')
        plt.clf()
        fig, ax = plt.subplots(4, 2, sharex=True, sharey=True, num='jam', gridspec_kw={'hspace': 0})

        if self.align == 'cyl':
            txt = [r"$\sigma_R$", r"$\sigma_z$", r"$\sigma_\phi$", r"$V^{\rm rms}_\phi$"]
        else:
            txt = [r"$\sigma_r$", r"$\sigma_\theta$", r"$\sigma_\phi$", r"$V^{\rm rms}_\phi$"]

        ax[0, 0].set_title("Data")
        ax[0, 1].set_title(f"JAM$_{{\\rm {self.align}}}$ Model")

        for j, (d, m, t) in enumerate(zip(self.data, np.sqrt(self.model.clip(0)), txt)):
            plt.sca(ax[j, 0])
            mn, mx = np.percentile(d, [1, 99])
            plot_velfield(self.Rbin, self.zbin, d, vmin=mn, vmax=mx, flux=self.flux, nodots=nodots)

            plt.sca(ax[j, 1])
            plot_velfield(self.Rbin, self.zbin, m, vmin=mn, vmax=mx, flux=self.flux,
                          colorbar=1, label=t, nodots=nodots)

##############################################################################



#!/usr/bin/env python
"""
    In this example/test I fit a one dimensional MGE to the density
    of a Hernquist (1990, ApJ, 356, 359; hereafter H90) model.
    I then compute the following quantities for a spherical model:
    1) The circular velocity with mge_vcirc;
    2) The projected sigma of an isotropic H90 model;
    3) The projected sigma of a fully tangential H90 model;
    4) The projected sigma of an ansiotropic H90 model with black hole;
    I use both jam_sph_rms and jam_axi_proj with align='cyl'/'sph';
    The solutions are compared with the analytic results by H90 and
    with the solution presented in Fig.4.20 of Binney & Tremaine (2008).

    - IMPORTANT: One needs the mge_fit_1d routine from the `mgefit`
      package available from http://purl.org/cappellari/software.

    V1.0.0: Michele Cappellari, Oxford, 28 November 2008
    V1.1.0: Included test with anisotropy and black holes.
      MC, Oxford, 17 June 2010
    V2.0.0: Translated from IDL into Python. MC, Oxford, 9 April 2014
    V2.0.1: Fixed RuntimeWarning. MC, Oxford, 17 March 2017
    V2.0.2: Changed imports for jam as a package. MC, Oxford, 17 April 2018
    V2.1.0: Use the new jampy.jam_axi_proj. MC, Oxford, 28 April 2021
    V2.2.0 Included align='sph' models and plot legend.
        MC, Oxford, 30 April 2021
    V2.2.1: Adapted for new jam_sph_proj. MC Oxford, 3 October 2022
    V2.3.0: Added Osipkov-Merritt test. MC, Oxford, 1 June 2023

"""

import numpy as np
import matplotlib.pyplot as plt

from mgefit.mge_fit_1d import mge_fit_1d
from jampy.mge_vcirc import mge_vcirc
from jampy.jam_axi_proj import jam_axi_proj
from jampy.jam_sph_proj import jam_sph_proj

#----------------------------------------------------------------------------

def jam_hernquist_model_example():

    M = 1e11          # Total mass in Solar Masses
    a = 1e3           # Break radius in pc
    distance = 16.5   # Assume Virgo distance in Mpc (Mei et al. 2007)
    pc = distance*np.pi/0.648 # Constant factor to convert arcsec --> pc
    G = 0.004301      # (km/s)^2 pc/Msun [6.674e-11 SI units (CODATA2018)]
    mbh = 0

    n = 300  # Number of values to sample the H90 profile for the fit
    r = np.logspace(-2.5, 2, n)*a       # logarithmically spaced radii in pc
    rho = M*a/(2*np.pi*r*(r + a)**3)    # Density in Msun/pc**3 (H90 equation 2)

    m = mge_fit_1d(r, rho, ngauss=16, plot=True)
    plt.pause(1)

    surf = m.sol[0]                 # Surface density in Msun/pc**2
    sigma = m.sol[1]/pc             # Gaussian dispersion in arcsec
    qObs = np.ones_like(surf)       # Assume spherical model
    inc = 90                        # Edge-on view
    npix = 100
    rad = np.geomspace(0.1, 100, npix) # desired output radii in arcsec (avoid R=0)
    r = rad*pc

    ################## Circular Velocity #################

    plt.clf()
    plt.subplot(211)
    plt.xlabel('R (arcsec)')
    plt.ylabel(r'$V_{\rm circ}$, $V_{\rm rms}$ (km/s)')
    plt.title('Reproduces projected analytic results by Hernquist (1990)')

    vcirc = mge_vcirc(surf, sigma, qObs, inc, mbh, distance, rad)
    plt.plot(rad, vcirc, 'C4', label=r'MGE V$_{\rm circ}$')

    # Compare with analytic result
    #
    vc = np.sqrt(G*M*r)/(r + a) # H90 equation (16)
    plt.plot(rad, vc, 'C0-.', label='Analytic')
    plt.text(30, 310, r'$V_{\rm circ}$')

    #################### Isotropic Vrms ###################

    # Spherical isotropic H90 model
    #
    jam = jam_sph_proj(surf, sigma, surf, sigma, mbh, distance, rad)
    sigp = jam.model
    plt.plot(rad, sigp, 'C1', label=r'JAM$^{\rm sph}_{\rm sph}$')

    # Axisymmetric isotropic model in the spherical limit.
    # This is plotted on the major axis, but the Vrms has circular symmetry
    #
    vrms = jam_axi_proj(surf, sigma, qObs, surf, sigma, qObs,
                        inc, mbh, distance, rad, rad*0, align='cyl').model
    plt.plot(rad, vrms, 'C2--', label=r'JAM$^{\rm axi}_{\rm cyl}$')

    vrms = jam_axi_proj(surf, sigma, qObs, surf, sigma, qObs,
                        inc, mbh, distance, rad, rad*0, align='sph').model
    plt.plot(rad, vrms, 'C3:', label=r'JAM$^{\rm axi}_{\rm sph}$')

    # Analytic surface brightness from H90
    #
    s = r/a
    w = s < 1
    xs = np.hstack([np.arccosh(1/s[w]) / np.sqrt(1 - s[w]**2),    # H90 eq. (33)
                    np.arccos(1/s[~w]) / np.sqrt(s[~w]**2 - 1)])  # H90 eq. (34)
    IR = M*((2 + s**2)*xs - 3) / (2*np.pi*a**2*(1 - s**2)**2)     # H90 eq. (32)

    # Projected second moments of isotropic model from H90
    #
    sigp = np.sqrt(G*M**2/(12*np.pi*a**3*IR) # H90 equation (41)
                 *(0.5/(1 - s**2)**3
                 *(-3*s**2*xs*(8*s**6 - 28*s**4 + 35*s**2 - 20)
                 - 24*s**6 + 68*s**4 - 65*s**2 + 6) - 6*np.pi*s))
    plt.plot(rad, sigp, 'C0-.')

    ################### Anisotropic Vrms ##################

    # Projected second moments for a H90 model with sigma_R=0.
    # This implies beta=-Infinity but I adopt as an approximation
    # below a large negative beta. This explains why te curves do not
    # overlap perfectly.
    #
    beta = np.full_like(surf, -20)
    jam = jam_sph_proj(surf, sigma, surf, sigma, mbh, distance, rad, beta=beta)
    sigp = jam.model
    plt.plot(rad, sigp, 'C1')

    # Axisymmetric anisotropic model in the spherical limit.
    # The spherical Vrms is the quadratic average of major
    # and minor axes of the axisymmetric model.
    #
    vrms_maj = jam_axi_proj(surf, sigma, qObs, surf, sigma, qObs,
                            inc, mbh, distance, rad, rad*0, beta=beta, align='cyl').model
    vrms_min = jam_axi_proj(surf, sigma, qObs, surf, sigma, qObs,
                            inc, mbh, distance, rad*0, rad, beta=beta, align='cyl').model
    vrms_r = jam_axi_proj(surf, sigma, qObs, surf, sigma, qObs,
                          inc, mbh, distance, rad, rad*0, beta=beta, align='sph').model
    plt.plot(rad, np.sqrt((vrms_maj**2 + vrms_min**2)/2), 'C2--')
    plt.plot(rad, vrms_r, 'C3:')

    # Projected second moments of fully tangential model from H90
    #
    sigp = np.sqrt(G*M**2*r**2/(2*np.pi*a**5*IR) # H90 equation (42)
                 *(1./(24*(1 - s**2)**4)
                 *(-xs*(24*s**8 - 108*s**6 + 189*s**4 - 120*s**2 + 120)
                 - 24*s**6 + 92*s**4 - 117*s**2 + 154) + 0.5*np.pi/s))
    plt.plot(rad, sigp, 'C0-.')

    plt.legend()

    ############### Osipkov-Merritt anisotropy variation ##############

    beta0 = 0       # Isotropic in the centre
    betainf = 1     # Fully radial at infinity
    ra = a/pc       # anisotropy radius in arcsec
    alpha = 2       # Osipkov-Merritt exponent
    beta = [ra, beta0, betainf, alpha]
    mbh = 0.005*M

    # The 'rani' keyword selects the specialized Osipkov-Merritt analytic solution
    sigp = jam_sph_proj(surf, sigma, surf, sigma, mbh, distance, rad, rani=ra).model

    # Here I use the generic logistic anisotropy but with Osipkov-Merritt values
    sigp1 = jam_sph_proj(surf, sigma, surf, sigma, mbh, distance, rad,
                         beta=beta, logistic=1).model

    # This is the axisymmetric model in the spherical limit and logistic anisotropy
    vrms_r = jam_axi_proj(surf, sigma, qObs, surf, sigma, qObs, inc, mbh,
                        distance, rad, rad*0, beta=beta, align='sph', logistic=1).model

    plt.semilogx(rad, sigp1, 'C1')
    plt.semilogx(rad, sigp, 'C0-.')
    plt.semilogx(rad, vrms_r, 'C3:')

    plt.annotate('Osipkov-Merritt+BH', xy=(0.8, 236), xycoords='data',
                 xytext=(0.5, 300), textcoords='data',
                 arrowprops=dict(facecolor='blue', ec='blue', arrowstyle="->"))
    plt.annotate('$V_{\\rm rms}(\\beta=-20\\approx-\infty)$\nfully tangential',
                 xy=(5, 160), xycoords='data', xytext=(2.6, 40), textcoords='data',
                 arrowprops=dict(facecolor='blue', ec='blue', arrowstyle="->"))
    plt.annotate('$V_{\\rm rms}(\\beta=0)$\nisotropic', xy=(28, 147),
                 xycoords='data', xytext=(13, 210), textcoords='data',
                 arrowprops=dict(facecolor='blue', ec='blue', arrowstyle="->"))
    plt.axvline(ra, ls='--', c='grey')
    plt.text(ra*1.02, 20, 'a')

    ############### Anisotropic models with Black Hole ###############

    # Reproduces Fig.4.20 of Binney J., & Tremaine S.D., 2008,
    # Galactic Dynamics, 2nd ed., Princeton University Press
    # See https://books.google.co.uk/books?id=6mF4CKxlbLsC&pg=PA352

    plt.subplot(212)
    plt.xlabel('R/a')
    plt.ylabel(r'$\sigma\, (G M / a)^{-1/2}$')
    plt.title('Reproduces Fig.4.20 bottom by Binney & Tremaine (2008)')

    cost = np.sqrt(G*M/a)
    rad = np.logspace(-2.3, 1, 50)*a/pc # desired output radii in arcsec
    bhs = np.array([0., 0.002, 0.004])*M
    betas = np.array([-0.51, 0, 0.51]) # Avoids singularity at beta=+/-0.5

    # Test the anisotropic jam_axi_proj(align='sph') with Black Hole in the spherical limit
    #
    for beta in betas:
        betaj = np.full_like(surf, beta)
        for bh in bhs:
            jam = jam_sph_proj(surf, sigma, surf, sigma, bh, distance, rad, beta=betaj)
            sigp = jam.model
            plt.semilogx(rad/a*pc, sigp/cost, 'C1')
            vrms = jam_axi_proj(surf, sigma, qObs, surf, sigma, qObs,
                                inc, bh, distance, rad, rad*0, beta=betaj, align='sph').model
            plt.semilogx(rad/a*pc, vrms/cost, 'C3:')

    # Test the isotropic jam_axi_proj(align='cyl') with Black Hole in the spherical limit
    #
    for bh in bhs:
        vrms = jam_axi_proj(surf, sigma, qObs, surf, sigma, qObs,
                            inc, bh, distance, rad, rad*0, align='cyl').model
        plt.semilogx(rad/a*pc, vrms/cost, 'C2--')

    plt.axvline(1, ls='--', c='grey')
    plt.text(1.02, 0.05, 'a')
    plt.axis([0.006, 10, 0, 0.6]) # x0, x1, y0, y1
    plt.tight_layout()

    plt.text(0.4, 0.5, r'$M_{\rm BH}=[0, 0.2\%, 0.4\%]\, M_*$')
    plt.text(0.03, 0.15, r'$\beta$=-0.5 (tangential)')
    plt.text(.03, 0.35, r'$\beta$=0 (isotropic)')
    plt.text(0.03, 0.5, r'$\beta$=0.5 (radial)')
    plt.pause(5)  # allow the plot to appear in certain situations

#----------------------------------------------------------------------------

if __name__ == '__main__':
    jam_hernquist_model_example()



#!/usr/bin/env python3
"""
    This example illustrates how to fit a galaxy model with dark matter halo
    using the JAM package in combination with the AdaMet Bayesian package.

    V1.0.0: Michele Cappellari, Oxford, 3 October 2022

"""
from os import path
import matplotlib.pyplot as plt
import numpy as np

# All packages below are available at https://pypi.org/user/micappe/

from plotbin.plot_velfield import plot_velfield
from plotbin.symmetrize_velfield import symmetrize_velfield

import jampy as jam_package
from jampy.jam_axi_proj import jam_axi_proj
from jampy.mge_half_light_isophote import mge_half_light_radius
from jampy.mge_radial_mass import mge_radial_mass

from adamet.adamet import adamet
from adamet.corner_plot import corner_plot

from mgefit.mge_fit_1d import mge_fit_1d

###############################################################################

def summary_plot(xbin, ybin, goodbins, rms, pars, lnprob, labels, bounds, kwargs):
    """
    Print the best fitting solution with uncertainties.
    Plot the final corner plot with the best fitting JAM model.

    """
    bestfit = pars[np.argmax(lnprob)]  # Best fitting parameters
    perc = np.percentile(pars, [15.86, 84.14], axis=0)  # 68% interval
    sig_bestfit = np.squeeze(np.diff(perc, axis=0)/2)   # half of interval (1sigma)

    print("\nBest-fitting parameters and 1sigma errors:")
    for label, best, sig in zip(labels, bestfit, sig_bestfit):
        print(f"   {label} = {best:#.4g} +/- {sig:#.2g}")

    # Produce final corner plot without trial values and with best fitting JAM
    plt.clf()
    corner_plot(pars, lnprob, labels=labels, extents=bounds, fignum=1)
    chi2 = jam_lnprob(bestfit, **kwargs)  # Compute model at best fit location

    dx = 0.24
    yfac = 0.87
    fig = plt.gcf()

    fig.add_axes([0.69, 0.99 - dx*yfac, dx, dx*yfac])  # left, bottom, xsize, ysize
    rms1 = rms.copy()
    rms1[goodbins] = symmetrize_velfield(xbin[goodbins], ybin[goodbins], rms[goodbins])
    vmin, vmax = np.percentile(rms1[goodbins], [0.5, 99.5])
    plot_velfield(xbin, ybin, rms1, vmin=vmin, vmax=vmax, cmap='viridis', linescolor='w',
                  colorbar=1, label=r"Data $V_{\rm rms}$ (km/s)", flux=jam_lnprob.flux_model)
    plt.tick_params(labelbottom=False)
    plt.ylabel('arcsec')

    fig.add_axes([0.69, 0.98 - 2*dx*yfac, dx, dx*yfac])  # left, bottom, xsize, ysize
    plot_velfield(xbin, ybin, jam_lnprob.rms_model, vmin=vmin, vmax=vmax, cmap='viridis', linescolor='w',
                  colorbar=1, label=r"Model $V_{\rm rms}$ (km/s)", flux=jam_lnprob.flux_model)
    plt.tick_params(labelbottom=False)
    plt.ylabel('arcsec')

###############################################################################

def dark_halo_mge(gamma, rbreak):
    """
    Returns the MGE parameters for a generalized NFW dark halo profile
    https://ui.adsabs.harvard.edu/abs/2001ApJ...555..504W
    - gamma is the inner logarithmic slope (gamma = -1 for NFW)
    - rbreak is the break radius in arcsec

    """
    # The fit is performed in log spaced radii from 1" to 10*rbreak
    n = 300     # Number of values to sample the gNFW profile for the MGE fit
    r = np.geomspace(1, rbreak*10, n)   # logarithmically spaced radii in arcsec
    rho = (r/rbreak)**gamma * (0.5 + 0.5*r/rbreak)**(-gamma - 3)  # rho=1 at r=rbreak
    m = mge_fit_1d(r, rho, ngauss=15, quiet=1, plot=0)

    surf_dm, sigma_dm = m.sol           # Peak surface density and sigma
    qobs_dm = np.ones_like(surf_dm)     # Assume spherical dark halo

    return surf_dm, sigma_dm, qobs_dm

###############################################################################

def total_mass_mge(surf_lum, sigma_lum, qobs_lum, gamma, rbreak, f_dm, inc):
    """
    Combine the MGE from a dark halo and the MGE from the stellar surface
    brightness in such a way to have a given dark matter fractions f_dm
    inside a sphere of radius one half-light radius reff

    """
    surf_dm, sigma_dm, qobs_dm = dark_halo_mge(gamma, rbreak)

    reff = mge_half_light_radius(surf_lum, sigma_lum, qobs_lum)[0]
    stars_lum_re = mge_radial_mass(surf_lum, sigma_lum, qobs_lum, inc, reff)
    dark_mass_re = mge_radial_mass(surf_dm, sigma_dm, qobs_dm, inc, reff)

    # Find the scale factor needed to satisfy the following definition
    # f_dm == dark_mass_re*scale/(stars_lum_re + dark_mass_re*scale)
    scale = (f_dm*stars_lum_re)/(dark_mass_re*(1 - f_dm))

    surf_pot = np.append(surf_lum, surf_dm*scale)   # Msun/pc**2. DM scaled so that f_DM(Re)=f_DM
    sigma_pot = np.append(sigma_lum, sigma_dm)      # Gaussian dispersion in arcsec
    qobs_pot = np.append(qobs_lum, qobs_dm)

    return surf_pot, sigma_pot, qobs_pot

###############################################################################

def jam_lnprob(pars, surf_lum=None, sigma_lum=None, qobs_lum=None, distance=None,
              xbin=None, ybin=None, sigmapsf=None, normpsf=None, goodbins=None,
              rms=None, erms=None, pixsize=None, plot=True):
    """
    Return the probability of the model, given the data, assuming constant priors

    """
    q, ratio, f_dm, lg_ml = pars

    # Sample inclination using min(q), with Equation (14) of Cappellari (2008)
    qmin = np.min(qobs_lum)
    inc = np.degrees(np.arctan2(np.sqrt(1 - qmin**2), np.sqrt(qmin**2 - q**2)))
    beta = np.full_like(qobs_lum, 1 - ratio**2)   # assume constant anisotropy

    # In this example I keep the halo slope and break radius fixed.
    # These parameters could be fitted from good and spatially-extended data.
    gamma = -1                  # Adopt fixed NFW inner halos slope
    rbreak = 20e3               # Adopt fixed halo break radius of 20 kpc (much larger than the data)
    pc = distance*np.pi/0.648   # Constant factor to convert arcsec --> pc
    rbreak /= pc                # Convert the break radius from pc --> arcsec
    mbh = 0                     # Ignore the central black hole
    surf_pot, sigma_pot, qobs_pot = total_mass_mge(surf_lum, sigma_lum, qobs_lum, gamma, rbreak, f_dm, inc)

    # Note: I multiply surf_pot by ml=10**lg_ml, while I set the keyword ml=1
    # Both the stellar and dark matter increase by ml and f_dm is unchanged
    surf_pot *= 10**lg_ml
    jam = jam_axi_proj(surf_lum, sigma_lum, qobs_lum, surf_pot, sigma_pot, qobs_pot,
                       inc, mbh, distance, xbin, ybin, plot=plot, pixsize=pixsize, quiet=1,
                       sigmapsf=sigmapsf, normpsf=normpsf, goodbins=goodbins, align='cyl',
                       beta=beta, data=rms, errors=erms, ml=1)

    # These two lines are just used for the final plot
    jam_lnprob.rms_model = jam.model
    jam_lnprob.flux_model = jam.flux

    resid = (rms[goodbins] - jam.model[goodbins])/erms[goodbins]
    chi2 = resid @ resid

    return -0.5*chi2   # ln(likelihood) + cost.

##############################################################################

def jam_dark_halo_adamet_example():

    # MGE model of the galaxy M32 from Table B1 of Cappellari et al. (2006)
    # http://adsabs.harvard.edu/abs/2006MNRAS.366.1126C
    surf_lum = 10**np.array([6.187, 5.774, 5.766, 5.613, 5.311, 4.774, 4.359, 4.087, 3.682, 3.316, 2.744, 1.618])
    sigma_lum = 10**np.array([-1.762, -1.143, -0.839, -0.438, -0.104, 0.232, 0.560, 0.835, 1.160, 1.414, 1.703, 2.249])
    qobs_lum = np.array([0.790, 0.741, 0.786, 0.757, 0.720, 0.724, 0.725, 0.743, 0.751, 0.838, 0.835, 0.720])

    # Read mock kinematics with realistic parameters and noise
    jam_dir = path.dirname(path.realpath(jam_package.__file__))
    filename = path.join(jam_dir, 'examples', 'jam_mock_kinematics_dark_halo.txt')
    xbin, ybin, rms, erms, flux = np.loadtxt(filename).T
    distance = 0.7   # M32 Distance

    # The following line tries to *approximately* account for systematic errors.
    # See beginning of Sec.6.1 of Mitzkus+17 for an explanation
    # https://ui.adsabs.harvard.edu/abs/2017MNRAS.464.4789M
    erms *= (2*rms.size)**0.25

    # Starting guess, e.g. from a previous least-squares fit
    q0 = 0.55               # Axial ratio of the flattest MGE Gaussian
    ratio0 = 0.9            # Anisotropy ratio sigma_z/sigma_R
    f_dm0 = 0.15            # Dark matter fraction inside a sphere of radius Re
    lg_ml0 = np.log10(1.4)  # I sample the M/L logarithmically

    pixsize = 2             # spaxel size in arcsec (before Voronoi binning)
    sigmapsf = 4/2.355      # sigma PSF in arcsec (=FWHM/2.355)
    normpsf = 1

    # I adjusted the fitting range below after an initial fit which
    # gave me an idea of some suitable ranges for the parameters
    qmin = np.min(qobs_lum)
    p0 = [q0, ratio0, f_dm0, lg_ml0]
    bounds = [[0.051, 0.5, 0, lg_ml0-0.2], [qmin, 1, 0.5, lg_ml0+0.2]]
    labels = [r"$q_{\rm min}$", r"$\sigma_z/\sigma_R$", r"$f_{\rm DM}$", r"$\lg(M_\ast/L)$"]

    goodbins = np.isfinite(xbin)  # Here I fit all bins

    # These parameters are passed to JAM
    kwargs = {'surf_lum': surf_lum, 'sigma_lum': sigma_lum, 'qobs_lum': qobs_lum,
              'distance': distance, 'xbin': xbin, 'ybin': ybin, 'sigmapsf': sigmapsf,
              'normpsf': normpsf, 'rms': rms, 'erms': erms, 'pixsize': pixsize,
              'goodbins': goodbins, 'plot': 0}

    # This is a rather small number of steps for illustration.
    # But the distribution has already qualitatively converged
    nstep = 1000
    sigpar = np.array([0.15, 0.15, 0.05, 0.05])  # crude estimate of uncertainties

    # The acceptance rate approaches the optimal theoretical value of 28% in 4-dim
    print("Started AdaMet please wait...")
    print("This example takes about 15 min on a 2GHz CPU")
    print("Progress is printed roughly every minute")
    pars, lnprob = adamet(jam_lnprob, p0, sigpar, bounds, nstep, fignum=1,
                          kwargs=kwargs, nprint=nstep/10, labels=labels, seed=2)

    summary_plot(xbin, ybin, goodbins, rms, pars, lnprob, labels, bounds, kwargs)

##############################################################################

if __name__ == '__main__':

    # This example takes about 15 min on a 2GHz CPU
    jam_dark_halo_adamet_example()



"""
    Copyright (C) 2016-2023, Michele Cappellari

    E-mail: michele.cappellari_at_physics.ox.ac.uk

    Updated versions of the software are available from my web page
    https://purl.org/cappellari/software

    This software is provided as is without any warranty whatsoever.
    Permission to use, for non-commercial purposes is granted.
    Permission to modify for personal or internal use is granted,
    provided this copyright and disclaimer are included unchanged
    at the beginning of the file. All other rights are reserved.
    In particular, redistribution of the code is not allowed.

    Changelog
    ---------

    V1.0.0: Michele Cappellari, Oxford, 6 June 2016
    +++++++++++++++++++++++++++++++++++++++++++++++

    - Written and tested as a separate procedure.

    V1.1.0: MC, Oxford, 17 August 2022
    ++++++++++++++++++++++++++++++++++

    - Included angular_diameter_distance and luminosity_distance.

    Vx.x.x: MC, Oxford
    ++++++++++++++++++

    - Additional changes are documented in the CHANGELOG of the JamPy package.
            
"""
import numpy as np

from jampy.quad1d import quad1d

###############################################################################

def integrand(z, omega_m, omega_k, omega_lam):
    """Equation (14) of Hogg-99 http://arxiv.org/abs/astro-ph/9905116v4"""

    ez = np.sqrt(omega_m*(1 + z)**3 + omega_k*(1 + z)**2 + omega_lam)

    return 1/ez

###############################################################################

def comoving_distance(z, h0=70, omega_m=0.3, omega_lam=None):
    """Calculate comoving distance D_C (in Mpc) of an object given its redshift"""
    
    if omega_lam is None:   # flat Universe
        omega_lam = 1 - omega_m

    # Equation (7) of Hogg-99
    omega_k = 1 - omega_m - omega_lam

    c = 299792.458                  #  speed of light in km/s
    dh = c/h0

    is_scalar = np.isscalar(z)
    z = np.atleast_1d(z)
    dc = np.zeros_like(z, dtype=float)
    for i, zi in enumerate(z):  # equation (15) of Hogg-99
        dc[i] = dh*quad1d(integrand, [0, zi], args=(omega_m, omega_k, omega_lam)).integ

    # Equation (16) of Hogg-99
    if omega_k > 0:
        dm = dh*np.sinh(np.sqrt(omega_k)*dc/dh)/np.sqrt(omega_k)
    elif omega_k < 0:
        dm = dh*np.sin(np.sqrt(np.abs(omega_k))*dc/dh)/np.sqrt(np.abs(omega_k))
    else:
        dm = dc

    if is_scalar:
        dm = dm.item()  # Make it a scalar if input was scalar

    return dm
 
##############################################################################

def angular_diameter_distance(z, **kwargs):
    """Calculate angular diameter distance D_A (in Mpc) of an object given its redshift"""

    return comoving_distance(z, **kwargs)/(1 + z)

##############################################################################

def luminosity_distance(z, **kwargs):
    """Calculate luminosity distance D_L (in Mpc) of an object given its redshift"""

    return comoving_distance(z, **kwargs)*(1 + z)

##############################################################################

if __name__ == '__main__':

    z = 3.0
    dm = comoving_distance(z)
    print(f"redshift: {z}\nComoving distance: {dm}\nLuminosity distance: {dm*(1 + z)}\n"
          f"Angular diameter distance: {dm/(1 + z)}")

    dm = comoving_distance(z, omega_m=0.4, omega_lam=0.3)
    print(f"redshift: {z}\nComoving distance: {dm}\nLuminosity distance: {dm*(1 + z)}\n"
          f"Angular diameter distance: {dm/(1 + z)}")



"""
##############################################################################

Copyright (C) 2022-2023, Michele Cappellari
E-mail: michele.cappellari_at_physics.ox.ac.uk

Updated versions of the software are available from my web page
http://purl.org/cappellari/software

If you have found this software useful for your research,
I would appreciate an acknowledgement to the use of the
"JAM modelling method of Cappellari (2008)"

        https://ui.adsabs.harvard.edu/abs/2008MNRAS.390...71C

This software is provided as is without any warranty whatsoever.
Permission to use, for non-commercial purposes is granted.
Permission to modify for personal or internal use is granted,
provided this copyright and disclaimer are included unchanged
at the beginning of the file. All other rights are reserved.
In particular, redistribution of the code is not allowed.

##############################################################################

MODIFICATION HISTORY:
    V1.0.0: Written and tested.
       Michele Cappellari, Oxford, 3 October 2022
    Vx.x.x: Additional changes are documented in the CHANGELOG of the JamPy package.

"""

from time import perf_counter as clock
from scipy import special
import numpy as np

from jampy.quad1d import quad1d

##############################################################################

def integrand(t, sigma_lum, sigma_pot, dens_lum, mass, Mbh, r, beta, logistic, component):
    """
    Implements the integrand of equation (40) of Cappellari (2008, MNRAS, 390, 71).
    Also allows for an analytic radial variation of the anisotropy.

    """ 
    # TANH Change of variables for Jeans r-integral (Sec.6.2 of Cappellari 2020, MNRAS, 494, 4819)
    # np.log([1e-6*mds, 3*mxs]) -> [r, inf]
    drdt = np.exp(t)
    r1 = r + drdt[:, None]    # Broadcast over radii and MGE parameters

    if logistic:
        # Variable anisotropy, same for all Gaussians
        ra, beta0, betainf, alpha = beta
        fun = (r1/r)**(2*beta0)
        fun *= ((1 + (r1/ra)**alpha)/(1 + (r/ra)**alpha))**(2*(betainf - beta0)/alpha)
        beta = beta0 + (betainf - beta0)/(1 + (ra/r)**alpha)
    else:
        # Constant anisotropy, different per Gaussian
        fun = (r1/r)**(2*beta)

    G = 0.004301  # (km/s)^2 pc/Msun [6.674e-11 SI units (CODATA2018)]
    h = r1/(np.sqrt(2)*sigma_pot)
    mass_r = Mbh + (mass*(special.erf(h) - 2/np.sqrt(np.pi)*h*np.exp(-h**2))).sum(1)    # eq.(49) of Cappellari (2008)
    nu_r = dens_lum*np.exp(-0.5*(r1/sigma_lum)**2)                                      # eq.(47) of Cappellari (2008)
    nuf = nu_r*fun/r1**2                                                                # eq.(40) of Cappellari (2008)

    if component == 'sig2r':
        nuf = nuf.sum(1)
    elif component == 'sig2th':
        nuf = (nuf*(1 - beta)).sum(1)

    integ = G*nuf*mass_r   # Vector of values computed at different radii

    return integ*drdt

##############################################################################

class jam_sph_intr:

    """
    PURPOSE
    -------

    This procedure calculates a prediction for the intrinsic second moments
    <v_r^2> and <v_th^2> in the radial and tangential directions for a
    spherically symmetric MGE model.
    It implements the solution of the anisotropic Jeans equations
    presented in equation (40) of `Cappellari (2008, MNRAS, 390, 71).
    <https://ui.adsabs.harvard.edu/abs/2008MNRAS.390...71C>`_

    CALLING SEQUENCE
    ----------------

    .. code-block:: python

        from jampy.jam_sph_intr import jam_sph_intr

        jam = jam_sph_intr(dens_lum, sigma_lum, dens_pot, sigma_pot, mbh, rad,
                           beta=None, epsrel=1e-2)
        sigma_r, sigma_th = np.sqrt(jam.model)

    INPUT PARAMETERS
    ----------------

    dens_lum:
        vector of length N containing the peak surface brightness of the
        MGE Gaussians describing the galaxy surface brightness in units of
        Lsun/pc^2 (solar luminosities per parsec^2).
    sigma_lum:
        vector of length N containing the dispersion in pc of
        the MGE Gaussians describing the galaxy surface brightness.
    surf_pot:
        vector of length M containing the peak value of the MGE Gaussians
        describing the galaxy surface density in units of Msun/pc^2 (solar
        masses per parsec^2). This is the MGE model from which the model
        potential is computed.

    sigma_pot:
        vector of length M containing the dispersion in pc of
        the MGE Gaussians describing the galaxy surface density.
    mbh:
        Mass of a nuclear supermassive black hole in solar masses.
    rad:
        Vector of length P with the (positive) radius from the galaxy center
        in pc at which one wants to compute the model predictions.

    Optional Keywords
    -----------------

    beta: array_like with shape (n,) or (4,)
        Radial anisotropy of the individual kinematic-tracer MGE Gaussians
        (Default: ``beta=np.zeros(n)``)::

            beta = 1 - (sigma_th/sigma_r)^2

        When ``logistic=True`` the procedure assumes::

            beta = [r_a, beta_0, beta_inf, alpha]

        and the anisotropy of the whole JAM model varies as a logistic
        function of the logarithmic spherical radius::

            beta(r) = beta_0 + (beta_inf - beta_0)/[1 + (r_a/r)^alpha]

        Here ``beta_0`` represents the anisotropy at ``r = 0``, ``beta_inf``
        is the anisotropy at ``r = inf`` and ``r_a`` is the anisotropy
        transition radius, with ``alpha`` controlling the sharpness of the
        transition. In the special case ``beta_0 = 0, beta_inf = 1, alpha = 2``
        the anisotropy variation reduces to the form by Osipkov & Merritt, but
        the extra parameters allow for much more realistic anisotropy profiles.
    epsrel: float, optional
        Relative error requested for the numerical quadrature
        (Default: ``epsrel=1e-2``)
    logistic: bool, optional
        When ``logistic=True``, JAM interprets the anisotropy parameter
        ``beta`` as defining a 4-parameters logistic function.
        See the documentation of the anisotropy keyword for details.
        (Default ``logistic=False``)

    Output Parameters
    -----------------

    Returned as attributes of the ``jam_sph_intr`` class.

    .flux: array_like  with shape (p,)
        Vector with the MGE luminosity density at each ``r`` location in ``Lsun/pc^3``.
    .model: array_like with shape (2, p)
        Contains ``[sig2r, sig2th]`` defined as follows:

        sig2r: array_like with shape (p,)
            Squared intrinsic dispersion in ``(km/s)^2`` along the r
            direction at each ``r`` location.

        sig2th: array_like with shape (p,)
            Squared intrinsic dispersion in ``(km/s)^2`` along the th
            direction at each ``r`` location.

    """

    def __init__(self, dens_lum, sigma_lum, dens_pot, sigma_pot, mbh, rad,
                 beta=None, epsrel=1e-2, logistic=False, quiet=False):

        if beta is None:
            beta = np.zeros_like(dens_lum)
        assert (dens_lum.size == sigma_lum.size) and \
               ((len(beta) == 4 and logistic) or (len(beta) == dens_lum.size)), \
            "The luminous MGE components and anisotropies do not match"
        assert len(dens_pot) == len(sigma_pot), 'surf_pot and sigma_pot must have the same length'

        t = clock()

        mass = dens_pot*(np.sqrt(2*np.pi)*sigma_pot)**3
        lim = np.log([1e-6*np.median(sigma_lum), 3*np.max(sigma_lum)])
        nu_sig2r, nu_sig2th = np.empty((2, rad.size))

        for j, rj in enumerate(rad):
            args = [sigma_lum, sigma_pot, dens_lum, mass, mbh, rj, beta, logistic]
            nu_sig2r[j], nu_sig2th[j] = \
                [quad1d(integrand, lim, epsrel=epsrel, args=args+[txt]).integ
                 for txt in ['sig2r', 'sig2th']]

        self.rad = rad
        self.flux = (dens_lum*np.exp(-0.5*(rad[:, None]/sigma_lum)**2)).sum(1)
        self.model = nu_sig2r/self.flux, nu_sig2th/self.flux

        if not quiet:
            print(f'jam_sph_intr elapsed time (sec): {clock() - t:.2f}')

##############################################################################



#!/usr/bin/env python3
"""
    This example illustrates how to fit the mass of a supermassive black hole
    using the JAM package in combination with the AdaMet Bayesian package.

    V1.0.0: Michele Cappellari, Oxford, 04 May 2018
    V1.1.0: Use the new jampy.jam_axi_proj. MC, Oxford, 28 April 2021

"""

from os import path
import matplotlib.pyplot as plt
import numpy as np

# All packages below are available at https://pypi.org/user/micappe/

from plotbin.plot_velfield import plot_velfield
from plotbin.symmetrize_velfield import symmetrize_velfield

import jampy as jam_package
from jampy.jam_axi_proj import jam_axi_proj

from adamet.adamet import adamet
from adamet.corner_plot import corner_plot

###############################################################################

def summary_plot(xbin, ybin, goodbins, rms, pars, lnprob, labels, bounds, kwargs):
    """
    Print the best fitting solution with errors.
    Plot the final corner plot with the best fitting JAM model.

    """
    bestfit = pars[np.argmax(lnprob)]  # Best fitting parameters
    perc = np.percentile(pars, [15.86, 84.14], axis=0)
    sig_bestfit = np.squeeze(np.diff(perc, axis=0)/2)   # half of 68% interval

    print("\nBest-fitting parameters and 1sigma errors:")
    for label, best, sig in zip(labels, bestfit, sig_bestfit):
        print(f"   {label} = {best:#.4g} +/- {sig:#.2g}")
        
    # Produce final corner plot wihout trial values and with best fitting JAM
    plt.clf()
    corner_plot(pars, lnprob, labels=labels, extents=bounds, fignum=1)
    chi2 = jam_lnprob(bestfit, **kwargs)  # Compute model at best fit location

    dx = 0.24
    yfac = 0.87
    fig = plt.gcf()

    fig.add_axes([0.69, 0.99 - dx*yfac, dx, dx*yfac])  # left, bottom, xsize, ysize
    rms1 = rms.copy()
    rms1[goodbins] = symmetrize_velfield(xbin[goodbins], ybin[goodbins], rms[goodbins])
    vmin, vmax = np.percentile(rms1[goodbins], [0.5, 99.5])
    plot_velfield(xbin, ybin, rms1, vmin=vmin, vmax=vmax, cmap='viridis', linescolor='w',
                  colorbar=1, label=r"Data $V_{\rm rms}$ (km/s)", flux=jam_lnprob.flux_model)
    plt.tick_params(labelbottom=False) 
    plt.ylabel('arcsec')

    fig.add_axes([0.69, 0.98 - 2*dx*yfac, dx, dx*yfac])  # left, bottom, xsize, ysize
    plot_velfield(xbin, ybin, jam_lnprob.rms_model, vmin=vmin, vmax=vmax, cmap='viridis', linescolor='w',
                  colorbar=1, label=r"Model $V_{\rm rms}$ (km/s)", flux=jam_lnprob.flux_model)
    plt.tick_params(labelbottom=False) 
    plt.ylabel('arcsec')

###############################################################################

def jam_lnprob(pars, surf_lum=None, sigma_lum=None, qobs_lum=None,
              surf_pot=None, sigma_pot=None, qobs_pot=None, dist=None, 
              xbin=None, ybin=None, sigmapsf=None, normpsf=None, goodbins=None, 
              rms=None, erms=None, pixsize=None, plot=True):
    """
    Return the probability of the model, given the data, assuming constant priors

    """
    q, ratio, mbh, lg_ml = pars
    
    # Sample inclination using min(q), with Equation (14) of Cappellari (2008)
    qmin = np.min(qobs_lum)
    inc = np.degrees(np.arctan2(np.sqrt(1 - qmin**2), np.sqrt(qmin**2 - q**2)))
    beta = np.full_like(qobs_lum, 1 - ratio**2)   # assume constant anisotropy

    # Note: surf_pot is multiplied by ml, while I set the keyword ml=1
    jam = jam_axi_proj(surf_lum, sigma_lum, qobs_lum, surf_pot*10**lg_ml, sigma_pot, qobs_pot,
                       inc, mbh, dist, xbin, ybin, plot=plot, pixsize=pixsize, quiet=1,
                       sigmapsf=sigmapsf, normpsf=normpsf, goodbins=goodbins, align='cyl',
                       beta=beta, data=rms, errors=erms, ml=1)

    # These two lines are just for the final plot
    jam_lnprob.rms_model = jam.model
    jam_lnprob.flux_model = jam.flux

    resid = (rms[goodbins] - jam.model[goodbins])/erms[goodbins]
    chi2 = resid @ resid

    return -0.5*chi2  # ln(likelihood) + cost.

##############################################################################

def jam_black_hole_adamet_example():

    # MGE model of M32 from Table B1 of Cappellari et al. (2006)
    # http://adsabs.harvard.edu/abs/2006MNRAS.366.1126C
    surf = 10**np.array([6.187, 5.774, 5.766, 5.613, 5.311, 4.774, 4.359, 4.087, 3.682, 3.316, 2.744, 1.618])
    sigma = 10**np.array([-1.762, -1.143, -0.839, -0.438, -0.104, 0.232, 0.560, 0.835, 1.160, 1.414, 1.703, 2.249])
    qObs = np.array([0.790, 0.741, 0.786, 0.757, 0.720, 0.724, 0.725, 0.743, 0.751, 0.838, 0.835, 0.720])

    # Read mock kinematics with realistic parameters and noise
    jam_dir = path.dirname(path.realpath(jam_package.__file__))
    filename = path.join(jam_dir, 'examples', 'jam_mock_kinematics_black_hole.txt')
    xbin, ybin, rms, erms, flux = np.loadtxt(filename).T
    distance = 0.7   # M32 Distance

    # Here assume mass follows light
    surf_lum = surf_pot = surf
    sigma_lum = sigma_pot = sigma
    qobs_lum = qobs_pot = qObs

    # Starting guess, e.g. from a previous least-squares fit
    q0 = 0.4
    bh0 = 2.5e6
    lg_ml0 = np.log10(1.4)
    ratio0 = 0.9    # anisotropy ratio sigma_z/sigma_R

    # Typical Adaptive-Optics PSF: narrow core + broad wings
    sigmapsf = [0.04, 0.4]  # sigma PSF in arcsec
    normpsf = [0.7, 0.3]
    pixsize = 0.05

    qmin = np.min(qObs)
    p0 = [q0, ratio0, bh0, lg_ml0]
    bounds = [[0.051, 0.5, bh0/1.3, lg_ml0-0.04], [qmin, 1.5, bh0*1.3, lg_ml0+0.04]]
    labels = [r"$q_{\rm min}$", r"$\sigma_z/\sigma_R$", r"$M_{\rm BH}$", r"$\lg(M/L)_{\rm tot}$"]

    goodbins = np.isfinite(xbin)  # Here we fit all bins

    # These parameters are passed to JAM
    kwargs = {'surf_lum': surf_lum, 'sigma_lum': sigma_lum, 'qobs_lum': qobs_lum,
              'surf_pot': surf_pot, 'sigma_pot': sigma_pot, 'qobs_pot': qobs_pot,
              'dist': distance, 'xbin': xbin, 'ybin': ybin, 'sigmapsf': sigmapsf,
              'normpsf': normpsf, 'rms': rms, 'erms': erms, 'pixsize': pixsize,
              'goodbins': goodbins, 'plot': 0}

    # This is a rather small number of steps for illustration.
    # But the distribution has already qualitatively converged
    nstep = 1000
    sigpar = np.array([0.1, 0.05, bh0*0.1, 0.01])  # crude estimate of uncertainties

    # The acceptance rate approaches the optimal theoretical value of 28% in 4-dim
    print("Started AdaMet please wait...")
    print("This example takes about 8 min on a 2GHz CPU")
    print("Progress is printed roughly every minute")
    pars, lnprob = adamet(jam_lnprob, p0, sigpar, bounds, nstep, fignum=1,
                          kwargs=kwargs, nprint=nstep/10, labels=labels, seed=3)

    summary_plot(xbin, ybin, goodbins, rms, pars, lnprob, labels, bounds, kwargs)
            
##############################################################################

if __name__ == '__main__':

    # This example takes about 8 min on a 2GHz CPU
    jam_black_hole_adamet_example()



"""
#############################################################################

Copyright (C) 2011-2017, Michele Cappellari
E-mail: michele.cappellari_at_physics.ox.ac.uk

Updated versions of the software are available from my web page
http://purl.org/cappellari/software

If you have found this software useful for your research,
I would appreciate an acknowledgement to the use of the
"JAM modelling package of Cappellari (2008)"

    https://ui.adsabs.harvard.edu/abs/2008MNRAS.390...71C
    
or an explicit reference to the equations given below.

This software is provided as is without any warranty whatsoever.
Permission to use, for non-commercial purposes is granted.
Permission to modify for personal or internal use is granted,
provided this copyright and disclaimer are included unchanged
at the beginning of the file. All other rights are reserved.
In particular, redistribution of the code is not allowed.

#############################################################################

MODIFICATION HISTORY:
    V1.0.0: Michele Cappellari, Oxford, 12 January 2011
    V2.0.0: Translated from IDL into Python. MC, Oxford, 21 February 2014
    V2.0.1: Cleaned up code and documentation. MC, Oxford, 23 October 2015
    V2.0.2: Check input parameters. MC, Oxford, 2 March 2016
    V2.0.3: Refer to published equation. MC, Oxford, 7 February 2017

"""

import numpy as np
from scipy.special import erf

def mge_radial_mass(surf, sigma, qobs, inc, rad, distance=1e-5):
    """
    Calculates the analytic mass or luminosity of an axisymmetric
    MGE within a sphere of a given radius, using equation (14) of
    `Mitzkus, Cappellari & Walcher, 2017, MNRAS, 464, 4789
    <https://ui.adsabs.harvard.edu/abs/2017MNRAS.464.4789M>`_

    :param surf: Peak surface brightness of each Gaussian in Lsun/pc**2
    :param sigma: Gaussian dispersion in arcsec
    :param qobs: Observed axial ratio of each Gaussian
    :param inc: inclination of the MGE in degrees
    :param rad: radius of the sphere in arcsec (scalar or vector)
    :param distance: Galaxy distance in Mpc
    :return: mass of the MGE within spheres of radii rad (same size as rad)
    """
    assert surf.size == sigma.size == qobs.size, "The MGE components do not match"

    rad = np.atleast_1d(rad)
    pc = distance*np.pi/0.648  # Constant factor to convert arcsec --> pc
    sigma = sigma*pc
    rad = rad*pc

    # total luminosity each Gaussian (eq.12 of Cappellari 2008, MNRAS, 390, 71)
    lum = 2.*np.pi*surf*qobs*sigma**2

    # Axisymmetric deprojection (eq.14 of Cappellari 2008, MNRAS, 390, 71)
    inc = np.radians(inc)
    qintr = qobs**2 - np.cos(inc)**2
    if np.any(qintr < 0):
        raise ValueError('Inclination too low q < 0')
    qintr = np.sqrt(qintr)/np.sin(inc)
    qintr = qintr.clip(0.001, 0.999) # Clip to avoid division by zero in mass_r
    
    # Compute analytic integral of MGE inside a sphere
    e = np.sqrt(1. - qintr**2)
    h = rad[:, None]/(np.sqrt(2.)*sigma*qintr)
    mass_r = (lum*(erf(h) - np.exp(-(h*qintr)**2)*erf(h*e)/e)).sum(1)
   
    return mass_r

#------------------------------------------------------------------------------
